#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include "tcfg.h"
#include "adapter.h"
#include "loops.h"

extern loop_t **loop_map;

/*
 * A cfg-adapter is an object allowing access to a sub-graph of the original CFG; corresponding to a loop. 
 * This sub-graph contains only nodes belonging to the loop; and inner loops are reduced to a single node
 * (which entry/exit edges of the loop are mapped to predecessor/successors for the node). Will
 * also create virtual entry/exit node to ensure we have only one entry & one exit. 
 * 
 * The original goal was to have an on-the-fly access without having to store anything in the adapter
 * but it was troublesome to implement, so in fact this is not terribly useful (it would have been 
 * better to just re-make another tcfg*), but anyway the code works currently... 
 *
 * graph = original tcfg
 * numnodes = total nodes in the cfg
 * numloops = loops in the cfg
 * loop_id = loop identifier of the loop associated with this cfg-adapter (put 0 for whole graph)
 */

adapter_p new_adapter(tcfg_node_t **graph, int numnodes, int numloops, int loop_id) {
	int i;
	int num_entries, num_exits;
	int edge_found;
	tcfg_edge_t *edge;
	adapter_p tmp = malloc(sizeof(adapter_t));
	tmp->graph = graph;
	tmp->adapted_nodes = malloc(sizeof(adnode_p)*(numnodes+numloops+2));
	memset(tmp->adapted_nodes, 0, sizeof(adnode_p)*(numnodes + numloops + 2));
	tmp->num_real_nodes = numnodes;
	tmp->real_entries = NULL;
	tmp->real_exits = NULL;
	tmp->bs_entries = bs_alloc(numnodes + numloops);
	tmp->bs_exits = bs_alloc(numnodes + numloops);
	bs_empty(tmp->bs_entries);
	bs_empty(tmp->bs_exits);

	// look for entry node(s)
	num_entries = 0;
	if (loop_id == 0) {
		// main graph: use real entries
		addToWorkList(&tmp->real_entries, graph[0]);
		for (i = 0; i < numnodes; i++) {
			if (graph[topo_tcfg[i]]->out == NULL) {
				addToWorkList(&tmp->real_exits, graph[topo_tcfg[i]]);
				//printf("REAL_EXIT: topo-order= %u id= %u (%x)\n", i, graph[topo_tcfg[i]]->id, graph[topo_tcfg[i]]);
				break;
			}
		}
//		printf("REAL_EXIT: id %u (%x)\n", numnodes - 1, graph[topo_tcfg[numnodes - 1]]);
		bs_add(tmp->bs_entries, 0);
		bs_add(tmp->bs_exits, topo_tcfg[numnodes - 1]);
		num_entries = 1;
		num_exits = 1;
	} else { 
		// loop: look for entry/exits edges
		for (i = 0; i < numnodes; i++) {
			if (graph[i] == NULL)
				continue;
			edge_found = 0;
			for (edge = graph[i]->in; edge; edge = edge->next_in) {
				if ((loop_map[edge->dst->id]->id == loop_id) && !isInner(loop_map[edge->src->id]->id, loop_id)) {
					edge_found = 1;
					break;
				}
			}
			if (edge_found != 0) {
				//printf("node %u is entry\n", i);
				bs_add(tmp->bs_entries, i);
				addToWorkList(&tmp->real_entries, graph[i]);
				num_entries++;
			}
		}
		for (i = 0; i < numnodes; i++) {
			if (graph[i] == NULL)
				continue;
			edge_found = 0;
			for (edge = graph[i]->out; edge; edge = edge->next_out) {
				if ((loop_map[edge->src->id]->id == loop_id) && !isInner(loop_map[edge->dst->id]->id, loop_id)) {
					edge_found = 1;
					break;
				}
			}
			if (edge_found != 0) {
				//printf("node %u is exit\n", i);
				bs_add(tmp->bs_exits, i);
				addToWorkList(&tmp->real_exits, graph[i]);
				num_exits++;
			}
		}
	}

		tmp->loop_id = loop_id;
		/*
		for (i = 0; i < numnodes; i++) {
			printf("loopid %u for node %u\n", loop_map[graph[i]->id]->id, graph[i]->id);
			if (loop_map[graph[i]->id]->id == loop_id)
					tmp->num_adnodes++;
		}
		*/
		tmp->num_loops = numloops;
	return tmp;
}

//return node ID
int adnode_id(adnode_p node) {
	return node->id;
}

//create adnode_p from real tcfg node (must not be called directly from outside)
adnode_p adapter_makenode(adapter_p ad, int type, int i) {
	adnode_p tmp = malloc(sizeof(adnode_t));
	tmp->id = i;
	tmp->ad = ad;
	tmp->type = type;
	if (type == ADNODE_TYPE_NODE) {
		// node type == normal node, corresponding to a physical node in the cfg
		tmp->node = ad->graph[i];
		tmp->loop_id = 0;
		assert(i < ad->num_real_nodes);
		assert(loop_map[ad->graph[i]->id]->id == ad->loop_id);
	} else if (type == ADNODE_TYPE_LOOP) {
		// virtual node representing a reduced loop
		tmp->node = NULL;
		tmp->loop_id =  i - ad->num_real_nodes;
		assert(i >= ad->num_real_nodes);
	} else if (type == ADNODE_TYPE_VIRTUAL_ENTRY) {
		// virtual entry node (constructed to be the predecessors of all loop entry edges)
		tmp->node = NULL;
	} else if (type == ADNODE_TYPE_VIRTUAL_EXIT) {
		// virtual exit node (constructed to be the successor of all loop exit edges)
		tmp->node = NULL;
	} else {
		abort();
	}

	ad->adapted_nodes[i] = tmp;
	return tmp;
}

/*
 * Get CFG-adapter node from number & type (type == ADNODE_TYPE_NODE/...LOOP/...VIRTUAL_ENTRY/...VIRTUAL_EXIT) 
 */
adnode_p adapter_getnode(adapter_p ad, int type, int i) {
	if (ad->adapted_nodes[i] == NULL) {
		return adapter_makenode(ad, type, i);
	} else return ad->adapted_nodes[i];
}

/*
 * Return max CFG-adapter node ID +1
 */ 
int adapter_numnodes(adapter_p ad) {
	return ad->num_real_nodes + ad->num_loops + 2;
}

// Get entry (real or virtual)
adnode_p adapter_exit(adapter_p ad) {
	return adapter_getnode(ad, ADNODE_TYPE_VIRTUAL_EXIT, ad->num_real_nodes + ad->num_loops + 1);
}

// Get exit node (real or virtual)
adnode_p adapter_entry(adapter_p ad) {
	return adapter_getnode(ad, ADNODE_TYPE_VIRTUAL_ENTRY, ad->num_real_nodes + ad->num_loops);
}

int adapter_is_entry(adnode_p node) {
	return (node->type == ADNODE_TYPE_VIRTUAL_ENTRY);
}

int adapter_is_exit(adnode_p node) {
	return (node->type == ADNODE_TYPE_VIRTUAL_EXIT);
}

// Return iterator for out-edges of a node
aditer_p adnode_outedge(adnode_p node) {
	int i;
	aditer_p iter = malloc(sizeof(aditer_t));
	iter->type = ADITER_OUTEDGE;
	iter->node = node;
	iter->ad = node->ad;
	iter->seen_loops = bs_alloc(node->ad->num_loops);
	iter->use_entryexit = 0;
	iter->cur_edge = NULL;

	if (node->type == ADNODE_TYPE_LOOP) {
		//current virtual loop node: the successors of the virtual node are the successors of the loop exit edges
                iter->cur_edge = NULL; 
		for (i = 0; i < iter->ad->num_real_nodes; i++) {
			if (isInner(loop_map[iter->ad->graph[i]->id]->id, node->loop_id)) {
				for (iter->cur_edge = iter->ad->graph[i]->out; iter->cur_edge; iter->cur_edge = iter->cur_edge->next_out) {
					if (!isInner(loop_map[iter->cur_edge->dst->id]->id,node->loop_id) && isInner(loop_map[iter->cur_edge->dst->id]->id, node->ad->loop_id)) {
						iter->cur_node = i;
						return iter;
					}
				}
			}
		}
	} else if (node->type == ADNODE_TYPE_NODE) {
		//normal node: the successors are the successors of the tcfg node that are not in an inner loop
		if (bs_test(iter->ad->bs_exits, node->id)) {
			iter->use_entryexit = 1;
		}
		for (iter->cur_edge = node->node->out; iter->cur_edge; iter->cur_edge = iter->cur_edge->next_out) {
			if (isInner(loop_map[iter->cur_edge->dst->id]->id, node->ad->loop_id)) {
				return iter;
			}
		}
	} else {
		//virtual entry node: the successors are the successors of the entry edges of the current loop
		iter->virt = (node->type == ADNODE_TYPE_VIRTUAL_ENTRY) ? node->ad->real_entries : NULL;
	}
	return iter;
}


// Return iterator for in-edges of a node
aditer_p adnode_inedge(adnode_p node) {
	int i;
	aditer_p iter = malloc(sizeof(aditer_t));
	iter->type = ADITER_INEDGE;
	iter->node = node;
	iter->ad = node->ad;
	iter->seen_loops = bs_alloc(node->ad->num_loops);
	iter->use_entryexit = 0;
	iter->cur_edge = NULL;

	if (node->type == ADNODE_TYPE_LOOP) {
                iter->cur_edge = NULL; 
		for (i = 0; i < iter->ad->num_real_nodes; i++) {
			if (isInner(loop_map[iter->ad->graph[i]->id]->id, node->loop_id)) {
				for (iter->cur_edge = iter->ad->graph[i]->in; iter->cur_edge; iter->cur_edge = iter->cur_edge->next_in) {
					if (!isInner(loop_map[iter->cur_edge->src->id]->id, node->loop_id) && isInner(loop_map[iter->cur_edge->dst->id]->id, node->ad->loop_id)) {
						iter->cur_node = i;
						return iter;
					}
				}
			}
		}
	} else if (node->type == ADNODE_TYPE_NODE) {
		if (bs_test(iter->ad->bs_entries, node->id)) {
			iter->use_entryexit = 1;
		}
		for (iter->cur_edge = node->node->in; iter->cur_edge; iter->cur_edge = iter->cur_edge->next_in) {
			if (isInner(loop_map[iter->cur_edge->src->id]->id, node->ad->loop_id)) {
				return iter;
			}
		}
	} else {
		iter->virt = (node->type == ADNODE_TYPE_VIRTUAL_ENTRY) ? NULL : node->ad->real_exits;
	}	
	return iter;
}

int adapter_exists_node(adapter_p ad, int node_idx) {
/*
	assert(node_idx < (ad->num_real_nodes + ad->num_loops + 2 ));
	assert(node_idx >= 0);
	*/
	if (node_idx >= (ad->num_real_nodes + ad->num_loops))
		return 1;
	if (node_idx >= ad->num_real_nodes) {
//	 printf("loopid %u loopaddr->id %u\n", node_idx - ad->num_real_nodes, loops[node_idx - ad->num_real_nodes]->id );
		return (loops[node_idx - ad->num_real_nodes]->parent == loops[ad->loop_id]);
        }
	return (ad->graph[node_idx] && (loop_map[ad->graph[node_idx]->id]->id == ad->loop_id));
}

aditer_p adapter_iter(adapter_p ad) {
	aditer_p iter = malloc(sizeof(aditer_t));
	iter->type = ADITER_GRAPH;
	iter->ad = ad;
	iter->cur_node = 0;
	iter->use_entryexit = 0;
	while (!adapter_exists_node(ad, iter->cur_node)) {
		iter->cur_node++;
	}
	iter->node = NULL;
	iter->seen_loops = bs_alloc(ad->num_loops);
	bs_empty(iter->seen_loops);
	return iter;
}

aditer_p adapter_reviter(adapter_p ad) {
	aditer_p iter = malloc(sizeof(aditer_t));
	iter->type = ADITER_GRAPH_REV;
	iter->ad = ad;
	iter->cur_node = ad->num_real_nodes + ad->num_loops + 2 - 1;
	while (!adapter_exists_node(ad, iter->cur_node)) {
		iter->cur_node--;
	}
	iter->node = NULL;
	iter->seen_loops = bs_alloc(ad->num_loops);
	bs_empty(iter->seen_loops);
	return iter;
}

int aditer_ended(aditer_p iter) {
	switch(iter->type) {
		case ADITER_INEDGE:
		case ADITER_OUTEDGE:
			if ((iter->node->type == ADNODE_TYPE_VIRTUAL_ENTRY) || (iter->node->type == ADNODE_TYPE_VIRTUAL_EXIT)) {
				return (iter->virt == NULL);
			}
			return ((iter->cur_edge == NULL) && (!iter->use_entryexit));
			break;
		case ADITER_GRAPH:
			return (iter->cur_node >= (iter->ad->num_real_nodes + iter->ad->num_loops + 2));
        case ADITER_GRAPH_REV:
        	return (iter->cur_node < 0);
                        
	}
	return (iter->node == NULL);
}

int aditer_seen_loop(aditer_p iter , int loop_id) {
	if (loop_id == iter->ad->loop_id)
		return 0;
	return (bs_test(iter->seen_loops, loop_id));
}
void aditer_mark_seen(aditer_p iter, int loop_id) {
	if (loop_id == iter->ad->loop_id)
		return;
	bs_add(iter->seen_loops, loop_id);
}



aditer_p aditer_next(aditer_p iter) {
	switch(iter->type) {
		case ADITER_INEDGE:
				if ((iter->cur_edge == NULL) && 
					(iter->use_entryexit == 1)) {
					iter->use_entryexit = 0;
					return iter;
				}
			if (iter->node->type == ADNODE_TYPE_VIRTUAL_EXIT) {
				iter->virt = iter->virt->next;
				return iter;
			}
		    do {
					iter->cur_edge = iter->cur_edge->next_in;
					if ((iter->node->type == ADNODE_TYPE_LOOP) && (iter->cur_edge == NULL) && (iter->cur_node < iter->ad->num_real_nodes)) {
						do {
							iter->cur_node++;
						} while ((iter->cur_node < iter->ad->num_real_nodes) && (!iter->ad->graph[iter->cur_node] || !isInner(loop_map[iter->ad->graph[iter->cur_node]->id]->id,  iter->node->loop_id) || (!iter->ad->graph[iter->cur_node]->in)));
						if (iter->cur_node < iter->ad->num_real_nodes)
							iter->cur_edge = iter->ad->graph[iter->cur_node]->in;
					}
				} while ((iter->cur_edge) && ((iter->node->type == ADNODE_TYPE_LOOP && isInner(loop_map[iter->cur_edge->src->id]->id, iter->node->loop_id)) || !isInner(loop_map[iter->cur_edge->src->id]->id, iter->ad->loop_id) || (aditer_seen_loop(iter, loop_map[iter->cur_edge->src->id]->id))));
				if ((iter->cur_edge) && (iter->ad->loop_id != loop_map[iter->cur_edge->src->id]->id))
					aditer_mark_seen(iter, loop_map[iter->cur_edge->src->id]->id);
				break;
		case ADITER_OUTEDGE:
				if ((iter->cur_edge == NULL) && 
					(iter->use_entryexit == 1)) {
					iter->use_entryexit = 0;
					return iter;
				}
			if (iter->node->type == ADNODE_TYPE_VIRTUAL_ENTRY) {
				iter->virt = iter->virt->next;
				return iter;
			}
				do {
					iter->cur_edge = iter->cur_edge->next_out;
					if ((iter->node->type == ADNODE_TYPE_LOOP) && (iter->cur_edge == NULL) && (iter->cur_node < iter->ad->num_real_nodes)) {
						do {
							iter->cur_node++;
						} while ((iter->cur_node < iter->ad->num_real_nodes) && (!iter->ad->graph[iter->cur_node] || !isInner(loop_map[iter->ad->graph[iter->cur_node]->id]->id,  iter->node->loop_id) || (!iter->ad->graph[iter->cur_node]->out)));
						if (iter->cur_node < iter->ad->num_real_nodes)
							iter->cur_edge = iter->ad->graph[iter->cur_node]->out;
					}
				} while ((iter->cur_edge) && ((iter->node->type == ADNODE_TYPE_LOOP && isInner(loop_map[iter->cur_edge->dst->id]->id, iter->node->loop_id)) || !isInner(loop_map[iter->cur_edge->dst->id]->id, iter->ad->loop_id) || (aditer_seen_loop(iter, loop_map[iter->cur_edge->dst->id]->id))));
				if ((iter->cur_edge) && (iter->ad->loop_id != loop_map[iter->cur_edge->dst->id]->id))
					aditer_mark_seen(iter, loop_map[iter->cur_edge->dst->id]->id);
				break;
		case ADITER_GRAPH:
				do {
					iter->cur_node++;
				} while ((iter->cur_node < (iter->ad->num_real_nodes + iter->ad->num_loops + 2)) && !adapter_exists_node(iter->ad, iter->cur_node));
				break;
		case ADITER_GRAPH_REV:
				do {
					iter->cur_node--;
				} while ((iter->cur_node >= 0) && !adapter_exists_node(iter->ad, iter->cur_node));
				break;
	}
	return iter;
}

adnode_p aditer_value(aditer_p iter) {
	tcfg_node_t *node;

	if ((iter->type == ADITER_GRAPH) || (iter->type == ADITER_GRAPH_REV)) {
           
		if (iter->cur_node < iter->ad->num_real_nodes) {
			return adapter_getnode(iter->ad, ADNODE_TYPE_NODE, iter->cur_node);
		} else if (iter->cur_node < (iter->ad->num_real_nodes + iter->ad->num_loops)) {
			return adapter_getnode(iter->ad, ADNODE_TYPE_LOOP, iter->cur_node);
		} else if (iter->cur_node < (iter->ad->num_real_nodes + iter->ad->num_loops + 1)) {
			return adapter_getnode(iter->ad, ADNODE_TYPE_VIRTUAL_ENTRY, iter->cur_node);
		} else {
			return adapter_getnode(iter->ad, ADNODE_TYPE_VIRTUAL_EXIT, iter->cur_node);
		}
	}

	if ((iter->node->type == ADNODE_TYPE_VIRTUAL_ENTRY) || (iter->node->type == ADNODE_TYPE_VIRTUAL_EXIT)){
		node = iter->virt->data;
	} else {
		if (iter->cur_edge == NULL) {
			assert(iter->use_entryexit);
			return ((iter->type == ADITER_INEDGE) ? adapter_entry(iter->ad) : adapter_exit(iter->ad));
		} else {
			node = (iter->type == ADITER_INEDGE) ? iter->cur_edge->src : iter->cur_edge->dst;
		}
	}
	assert(node);

	if (loop_map[node->id]->id == iter->ad->loop_id)
		return adapter_getnode(iter->ad, ADNODE_TYPE_NODE, node->id);
	return adapter_getnode(iter->ad, ADNODE_TYPE_LOOP, loop_map[node->id]->id + iter->ad->num_real_nodes);
}


