/*******************************************************************************
 *
 * Chronos: A Timing Analyzer for Embedded Software
 * =============================================================================
 * http://www.comp.nus.edu.sg/~rpembed/chronos/
 *
 * Infeasible path analysis for Chronos
 * Vivy Suhendra
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * 03/2007 symexec.c
 *
 * Version update
 * v1.1: 
 *   _ perform sym.exec to identify constant value in register first
 *     then detect induction val., to handle: add $1 $1 $2, where $2 = const
 *   _ perform simple memory modeling with lw,sw; to deal with register spill
 *   _ parse disassembly file using C-string rather than shell call -> faster
 * v1.0:
 *   _ add induction value detection of register value
 *   _ add merge register value from predecessors to support inter-block
 *     deritree construction (still intra-procedures)
 *   _ add other things necessary to perform register expansion for RTSS2010
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cfg.h"
#include "common.h"
#include "loops.h"
#include "infeasible.h"

#define MIN_VAL (0-0xffff)

/*** MEMORY MODELING FRAMEWORK ***/
void    printMemList(FILE* fp, worklist_p);
mem_p   createMem(addr_t instAddr, reg_t writeAddr, reg_t regValue);
int     copyMem(mem_p dst, mem_p src);
void    clrMem(worklist_p memList);
boolean_t    loadMem(worklist_p memList, reg_t writeAddr, reg_t *regValue);
int     storeMem(worklist_p *memList, addr_t instAddr, 
                    reg_t writeAddr, reg_t regValue);
int     copyMemList(worklist_p *dstL, worklist_p srcL);
int     memListEq(worklist_p list1, worklist_p list2);
int     mergeMemList(worklist_p *dstList, worklist_p srcList);

/*** INDUCTION VALUE DETECTION ***/
biv_p   createBIV(reg_t *regList, insn_t *insn);
static int  regRedefined(char *regName, insn_t *insn);
static int  checkBivStride(reg_t *regList, insn_t *insn);
static void detectBIV();

/*** SYMBOLIC EXECUTION FRAMEWORK ***/
int     execute();

static int  doSymExec();
#ifdef _INF
static int  joinBrach(inf_proc_t *iproc, inf_node_t *ib, int lpHeaderId);
static int  regDerive(inf_node_t *ib,insn_t *insn,int ignore,char *jal,int dbg);
static void analyzeBlock(inf_node_t *ib);
static void readSymExecResult(inf_node_t *ib);
#else
static void readSymExecResult(tcfg_node_t *bbi);
#endif
static void initAll();
static void freeAll();

int     cntExec;
int     conflictDetectFlag; /*Global: if do conflicting pair detection*/
int     traceInductionFlag; /*Global: if trace induction value during sym.exec*/
int     traceMemFlag;   /*Global: if perform memory modeling during sym.exec*/ 

FILE* dbgExec; /*to debug the convergence*/

extern prog_t prog;
extern loop_t **loops;
extern int num_tcfg_loops;

extern int sp_init_value; //cleekee: initial stack pointer value

/*** utility functions ***/
int getAddrD(dat_inst_t *d_inst) {
    insn_t      *insn;
    insn = (insn_t*)(d_inst->insn);
    return hexValue(insn->addr);
}

void printDataRef(FILE *fp,dat_inst_t *d_inst) {
    int i;
    insn_t      *insn;
    //de_inst_t   *inst;
    char        *deritree;

    insn = (insn_t*)(d_inst->insn);
    //inst = d_inst->inst;
    //deritree = d_inst->deritree;
    if (fp) {
        fprintf(fp, "\n\nDataRef: 0x%s %s %s %s %s\n", 
                    insn->addr, insn->op, insn->r1, insn->r2, insn->r3 );
        printExpr(fp,&(d_inst->addrExpr));
    }
}

int isMemAccess(char* isa_name)
{
	return (isLoadInst(isa_name) || isStoreInst(isa_name));
}

int isLoadInst(char *isa_name) {
#ifdef ARCH_ARM
	if (strcmp(isa_name, "ldr") == 0 || strcmp(isa_name, "ldrb") == 0)
		return 1;
	else
		return 0;
#else
    if(! strcmp(isa_name, "lw") || !strcmp(isa_name, "lh") 
            || !strcmp(isa_name, "lb") || !strcmp(isa_name, "lwl")
            || !strcmp(isa_name, "lwr") || !strcmp(isa_name, "lhu") 
            || !strcmp(isa_name, "lbu") 
            || !strcmp(isa_name, "l.s") || !strcmp(isa_name, "l.d")) {
        return 1;
    }
    else return 0;
#endif
}

int isStoreInst(char *isa_name) {
#ifdef ARCH_ARM
	if (strcmp(isa_name, "str") == 0 || strcmp(isa_name, "strb") == 0)
		return 1;
	else
		return 0;
#else
	if(! strcmp(isa_name, "sw") || !strcmp(isa_name, "sh") 
	   || !strcmp(isa_name, "sb") || !strcmp(isa_name, "swl")
		|| !strcmp(isa_name, "swr") || !strcmp(isa_name, "s.s")
		|| !strcmp(isa_name, "s.d")) {
		return 1;  
    }
    else return 0;
#endif
}

loop_t* getIbLoop(inf_node_t *ib) {
    int dbg = 0;
    loop_t *loop;
    if (ib->loop_id==-1) return loops[0];
    loop = loops[inf_loops[ib->loop_id].loop_id]; 
    return loop;
}

void copyInst(insn_t *dsrc, insn_t *ddes) {
    strcpy(ddes->addr,dsrc->addr);
    strcpy(ddes->op,dsrc->op);
    strcpy(ddes->r1,dsrc->r1);
    strcpy(ddes->r2,dsrc->r2);
    strcpy(ddes->r3,dsrc->r3);
}

static int checkBivStride(reg_t *regList, insn_t* insn) {
    char str[128];
    int  k;
    int  regPos1,regPos2,regPos3;
    int  flag;

    regPos1 = findReg(regList, insn->r1);
    regPos2 = findReg(regList, insn->r2);
    if (regPos1==-1 || regPos2==-1) return INFINITY;

    k = INFINITY;
#ifndef ARCH_ARM
    if (regPos2==28 || regPos2==29) return INFINITY;/*ignore special reg*/
#endif
    if (strcmp(regList[regPos1].name,regList[regPos2].name)!=0) return INFINITY;
	
#ifdef ARCH_ARM
    if (streq(insn->op,"add")) {
        regPos3 = findReg(regList,insn->r3);
		if (regPos3 == -1) {
			/* r1 = r2 + k */
			k = atoi(insn->r3);
		} else if (regList[regPos3].t != VALUE_CONST) {
			/* r1 = r1 + r3 */
			return INFINITY;/*r3 not constant*/ 
        } else {
			/* r1 = r1 + r3 */
            k = regList[regPos3].val;
            return k;
        }
    }
#else
    if (streq(insn->op,"addu")) {
        /* r1 = r1 + r3 */
        regPos3 = findReg(regList,insn->r3);
        if (regList[regPos3].t!=VALUE_CONST) return INFINITY;/*r3 not constant*/ 
        else {
            k = regList[regPos3].val;
            return k;
        }
    }
    else if (streq(insn->op,"addi") || streq(insn->op,"addiu")) {
        /* r1 = r2 + k */
        k = atoi(insn->r3);
    }
#endif

    return k;
}

biv_p createBIV(reg_t *regList, insn_t *insn) {
    biv_p biVar;
    biVar = calloc(1,sizeof(biv_s));
    //sprintf(biVar->regName,"%s",insn->r1);
    strcpy(biVar->opr,insn->op);
    biVar->insn = insn;
    biVar->stride = checkBivStride(regList, insn);
    initReg( &(biVar->initVal));
    biVar->flag = 0;
    strcpy(biVar->regName,insn->r1);
    return biVar;
}

#ifdef _INF
int saved2Mem(inf_node_t *node, biv_p biv) {
    int dbg = 0;
    int k;
    insn_t *insn;
    char *bivAddr;
    //check if biv saved to mem before it is redefined
    //return 1 if (i) biv is saved, or (ii) biv is redefined. return 0->unknown
    //because saved biv detection is unaffected by loop scope,
    //it will appear in outermost loop where it belongs (not the loop using it)
    bivAddr = ((insn_t*)(biv->insn))->addr;
    if (dbg) {
    }
    for (k=0; k<node->num_insn; k++) {
        insn = &(node->insnlist[k]);
        if (isStoreInst(insn->op) && strcmp(biv->regName,insn->r1)==0
                && hexValue(insn->addr)>hexValue(bivAddr)) {
            biv->flag |= BIV_SAVED;
            if (dbg) {
                fprintf(dbgExec,"\nSaved: insn %s |%s(%s,0,%d)",
                        insn->addr,((insn_t*)(biv->insn))->addr, 
                        biv->regName,biv->stride);
            }
            return 1;
        }
        //(1):biv is redefined by this insn
        else if (strcmp(bivAddr,insn->addr)!=0
                && (biv->flag & BIV_SAVED)==0
                && hexValue(insn->addr)>hexValue(bivAddr)
                && regRedefined(biv->regName, insn)) {
            if (dbg) {
                fprintf(dbgExec,"\nRedefined: insn %s |%s(%s,0,%d)",
                        insn->addr,bivAddr,biv->regName,biv->stride);
                fflush(dbgExec);
            }
            return 1;
            #if 0
            if (prvNode) prvNode->next = bivNode->next;
            else biv_list = bivNode->next;
            freeBIV(&biv); free(bivNode);
            if (prvNode) bivNode = prvNode->next;
            else bivNode = biv_list;
            #endif
        }
    }
    return 0;
}
#else /* sudiptac: changed to handle tcfg node structure */
int saved2Mem(tcfg_node_t *node, biv_p biv) {
    int dbg = 0;
    int k;
    insn_t *insn;
    char *bivAddr;
    //check if biv saved to mem before it is redefined
    //return 1 if (i) biv is saved, or (ii) biv is redefined. 
		//return 0->unknown because saved biv detection is 
		//unaffected by loop scope, it will appear in outermost 
		//loop where it belongs (not the loop using it)
    bivAddr = ((insn_t*)(biv->insn))->addr;
    
		for (k = 0; k < node->bb->num_inst; k++) {
        insn = &(node->insnlist[k]);
        if (isStoreInst(insn->op) && strcmp(biv->regName,insn->r1)==0
                && hexValue(insn->addr)>hexValue(bivAddr)) {
            biv->flag |= BIV_SAVED;
            if (dbg) {
                fprintf(dbgExec,"\nSaved: insn %s |%s(%s,0,%d)",
                        insn->addr,((insn_t*)(biv->insn))->addr, 
                        biv->regName,biv->stride);
            }
            return 1;
        }
        //(1):biv is redefined by this insn
        else if (strcmp(bivAddr,insn->addr)!=0
                && (biv->flag & BIV_SAVED)==0
                && hexValue(insn->addr)>hexValue(bivAddr)
                && regRedefined(biv->regName, insn)) {
            if (dbg) {
                fprintf(dbgExec,"\nRedefined: insn %s |%s(%s,0,%d)",
                        insn->addr,bivAddr,biv->regName,biv->stride);
                fflush(dbgExec);
            }
            return 1;
        }
    }
    return 0;
}
#endif

/*** Loop induction variable detection ***/
/* sudiptac: modify to reflect the new position 
 * of "insnlist" in tcfg node structure */
#ifdef _INF
static void detectBIV() {
    int     dbg = 0; /*dbg flag*/
    int     i,j,k,h;
    char    *bivAddr;
    
    biv_p   biv;
    int     str;
    worklist_p biv_list, bivNode, prvNode;
    insn_t  *insn;

    inf_node_t *cur;
    loop_t  *curLp;
    proc_t  *p;
    int     lpHeadId, lpTailId, pId, lpId;

    reg_t   *regList;

    for (i = num_tcfg_loops - 1; i > 0; i--) {
        curLp = loops[i];
        if (curLp == NULL) continue;

        biv_list = NULL;
        p = curLp->head->bb->proc;
        if (curLp->head) 
					lpHeadId = curLp->head->bb->id;
        else 
					lpHeadId = 0;
        if (curLp->tail) 
					lpTailId = curLp->tail->bb->id;
        else 
					lpTailId = p->num_bb-1;
        if (dbg) {
            fprintf(dbgExec,"\nLoop %d head (%d:%d) tail (%d:%d)",
                curLp->id, p->id, lpHeadId, p->id, lpTailId);
            fflush(dbgExec);
        }

        //detect possible biv within the loop
        for (j = lpHeadId; j <= lpTailId; j++) {
            /*scan through inst, detect basic induction var*/
            cur = &(inf_procs[p->id].inf_cfg[j]);
            regList = (reg_t*)(cur->regListIn);
            for (k=0; k<cur->num_insn; k++) {
                insn = &(cur->insnlist[k]);
                if (insn->flag==BIV_INST) continue;//induction inst of outer lp
                str = checkBivStride(regList,insn);
                if (str != INFINITY) {
                    if (dbg) {
                        fprintf(dbgExec,"\nBIV: insn: %s (%s,_,%d)",
                            insn->addr,insn->r1,str);fflush(dbgExec);
                    }
                    //insn->flag = BIV_INST;
                    biv = createBIV(regList,insn);
                    //check whether a biv is saved before it is redefined
                    for (h=j; h<= lpTailId; h++) {
                        if (saved2Mem(inf_procs[p->id].inf_cfg+h,biv)) break;
                    }
                    if (0) {
                        fprintf(dbgExec,"\nSaved BIV: insn: %s - ",insn->addr);
                        printBIV(dbgExec,biv);
                        fflush(dbgExec);
                    }
                    //biv->lp = curLp;
                    addToWorkList(&biv_list, biv);
                }
            }//end insn[k]
        }//end bb[j]

        for (j=lpHeadId; j<=lpTailId; j++) {
            cur = &(inf_procs[p->id].inf_cfg[j]);
            regList = (reg_t*)(cur->regListIn);
            for (k=0; k<cur->num_insn; k++) {
                insn = &(cur->insnlist[k]);
                bivNode = biv_list; prvNode = NULL;
                while (bivNode) {
                    biv = bivNode->data;
                    bivAddr = ((insn_t*)(biv->insn))->addr;
                    if (biv==NULL) {bivNode=bivNode->next; continue;}
                    //check whether a possible biv is not a real biv because it is redefined (before it is saved to memory)
                    if (strcmp(bivAddr,insn->addr)!=0
                            && (biv->flag & BIV_SAVED)==0
                            && regRedefined(biv->regName, insn)) {
                        if (dbg) {
                            fprintf(dbgExec,"\nRedefined: insn %s |%s(%s,0,%d)",
                                    insn->addr,bivAddr,biv->regName,biv->stride);
                            fflush(dbgExec);
                        }
                        if (prvNode) prvNode->next = bivNode->next;
                        else biv_list = bivNode->next;
                        freeBIV(&biv); free(bivNode);
                        if (prvNode) bivNode = prvNode->next;
                        else bivNode = biv_list;
                    }
                    else {//not redefined by this insn
                        prvNode = bivNode;
                        bivNode = bivNode->next;
                    }
                }//end bivNode
            }//end insn[k]
        }//end bb[j]

        //record real biv
        curLp->biv_list = biv_list;
        for (bivNode = biv_list; bivNode; bivNode=bivNode->next) {
            biv = bivNode->data;
            if (biv) ((insn_t*)(biv->insn))->flag |= BIV_INST;
        }
        if (dbg) {
            fprintf(dbgExec,"\n BIV list of L%d ",curLp->id);
            for (bivNode = biv_list; bivNode; bivNode = bivNode->next) {
                biv = (biv_p) bivNode->data;
                bivAddr = ((insn_t*)(biv->insn))->addr;
                lpId = curLp->id;
                //lpId = ((loop_t*)(biv->lp))->id;
                fprintf(dbgExec,"\n\t %s %s L%d",bivAddr,biv->regName,lpId);
                printBIV(dbgExec,biv);
            }
        }
    }//end loop[i]
}
#else
static void detectBIV() {
    int     dbg = 0; /*dbg flag*/
    int     i,j,k,h;
    char    *bivAddr;
    
    biv_p   biv;
    int     str;
    worklist_p biv_list, bivNode, prvNode;
    insn_t  *insn;

    tcfg_node_t *cur;
    loop_t  *curLp;
    proc_t  *p;
    int     lpHeadId, lpTailId, pId, lpId;

    reg_t   *regList;

    for (i = num_tcfg_loops - 1; i > 0; i--) {
        curLp = loops[i];
        if (curLp == NULL) continue;

        biv_list = NULL;
        p = curLp->head->bb->proc;
        if (curLp->head) 
					lpHeadId = curLp->head->id;
        else 
					lpHeadId = 0;
        if (curLp->tail) 
					lpTailId = curLp->tail->id;
        else 
					lpTailId = num_tcfg_nodes - 1;
        
		if (dbg) {
            fprintf(dbgExec,"\nLoop %d head (%d:%d) tail (%d:%d)",
                curLp->id, p->id, lpHeadId, p->id, lpTailId);
            fflush(dbgExec);
        }

        /* detect possible biv within the loop */
		for(j=0; j<num_tcfg_nodes; j++) {
			if (loop_map[j] != curLp) continue;
			
            /*scan through inst, detect basic induction var*/
            cur = tcfg[j];
            regList = (reg_t*)(cur->regListIn);
            
						for (k=0; k < tcfg[j]->bb->num_inst; k++) {
                insn = &(cur->insnlist[k]);
                if (insn->flag==BIV_INST) continue; /* induction inst of outer lp */
                str = checkBivStride(regList,insn);
                if (str != INFINITY) {
                    if (dbg) {
                        fprintf(dbgExec,"\nBIV: insn: %s (%s,_,%d)",
                            insn->addr,insn->r1,str);fflush(dbgExec);
                    }
                    /* insn->flag = BIV_INST; */
                    biv = createBIV(regList,insn);
                    /* check whether a biv is saved before it is redefined */
                    for (h = j; h <= lpTailId; h++) {
						/* sudiptac: changed implementation to handle tcfg node structure */
                        if (saved2Mem(tcfg[h], biv)) break;
                    }
                    if (dbg) {
                        fprintf(dbgExec, "\nSaved BIV: insn: %s - ", insn->addr);
                        printBIV(dbgExec, biv);
                        fflush(dbgExec);
                    }
                    //biv->lp = curLp;
                    addToWorkList(&biv_list, biv);
                }
            }//end insn[k]
        }//end bb[j]

		for(j=0; j<num_tcfg_nodes; j++) {
			if (loop_map[j] != curLp) continue;
			
            cur = tcfg[j];
            regList = (reg_t*)(cur->regListIn);
            for (k = 0; k < cur->bb->num_inst; k++) {
                insn = &(cur->insnlist[k]);
                bivNode = biv_list; 
								prvNode = NULL;
                while (bivNode) {
                    biv = bivNode->data;
                    bivAddr = ((insn_t*)(biv->insn))->addr;
                    if (biv == NULL) {bivNode = bivNode->next; continue;}
                    /* check whether a possible biv is not a real biv 
										 * because it is redefined (before it is saved to 
										 * memory) */
                    if (strcmp(bivAddr,insn->addr)!=0
                            && (biv->flag & BIV_SAVED)==0
                            && regRedefined(biv->regName, insn)) {
                        if (dbg) {
                            fprintf(dbgExec,"\nRedefined: insn %s |%s(%s,0,%d)",
                                    insn->addr,bivAddr,biv->regName,biv->stride);
                            fflush(dbgExec);
                        }
                        if (prvNode) prvNode->next = bivNode->next;
                        else biv_list = bivNode->next;
                        freeBIV(&biv); free(bivNode);
                        if (prvNode) bivNode = prvNode->next;
                        else bivNode = biv_list;
                    }
                    else {/* not redefined by this insn */
                        prvNode = bivNode;
                        bivNode = bivNode->next;
                    }
                }/* end bivNode */
            }/* end insn[k] */
        }/* end bb[j] */

        //record real biv
        curLp->biv_list = biv_list;
        for (bivNode = biv_list; bivNode; bivNode=bivNode->next) {
            biv = bivNode->data;
            if (biv) ((insn_t*)(biv->insn))->flag |= BIV_INST;
        }
        if (dbg) {
            fprintf(dbgExec,"\n BIV list of L%d ",curLp->id);
            for (bivNode = biv_list; bivNode; bivNode = bivNode->next) {
                biv = (biv_p) bivNode->data;
                bivAddr = ((insn_t*)(biv->insn))->addr;
                lpId = curLp->id;
                //lpId = ((loop_t*)(biv->lp))->id;
                fprintf(dbgExec,"\n\t %s %s L%d",bivAddr,biv->regName,lpId);
                printBIV(dbgExec,biv);
            }
        }
    }//end loop[i]
}
#endif

/*** Functions for conflicting pair detection ***/
#if 0 /*HBK: this should belong to conflict.c or infeasible.c, not support now*/
int addAssign( char deritree[], cfg_node_t *bb, int lineno, int rhs, char rhs_var ) {
    int dbg = 0;
    int i;
    assign_t *assg;
    inf_node_t *ib;

    if (conflictDetectFlag==0) return;/*not do conflicting pair detection yet*/
    ib = &(inf_procs[bb->proc->id].inf_cfg[bb->id]);

    // check if there is previous assignment to the same memory address in the same block
    for( i = 0; i < ib->num_assign; i++ ) {
        assg = ib->assignlist[i];

        if( streq( assg->deritree, deritree )) {
            // overwrite this assignment
            assg->lineno  = lineno;
            assg->rhs     = rhs;
            assg->rhs_var = rhs_var;

            if( dbg & 0 ) { printf( "OVR " ); printAssign( assg, 0 ); }
            return 1;
        }
    }

    // else add new
    assg = (assign_t*) malloc( sizeof(assign_t) );
    strcpy( assg->deritree, deritree );
    assg->rhs           = rhs;
    assg->rhs_var       = rhs_var;
    assg->bb            = bb;
    assg->lineno        = lineno;

    ib->num_assign++;

    i = ib->num_assign;
    ib->assignlist = (assign_t**) realloc( ib->assignlist, i * sizeof(assign_t*) );
    ib->assignlist[i-1] = assg;

    if( dbg & 0 ) { printf( "NEW " ); printAssign( assg, 0 ); }
    return 0;
}

int addBranch( char deritree[], cfg_node_t *bb, int rhs, char rhs_var, char jump_cond ) {
    int dbg = 0;
    int i;
    branch_t *br;

    int numbb = prog.procs[bb->proc->id].num_bb;

    if (conflictDetectFlag==0) return;/*not do conflicting pair detection yet*/
    br = (branch_t*) malloc( sizeof(branch_t) );
    strcpy( br->deritree, deritree );
    br->rhs              = rhs;
    br->rhs_var          = rhs_var;
    br->jump_cond        = jump_cond;
    br->bb               = bb;

    br->num_BA_conflicts = 0;
    br->BA_conflict_list = NULL;
    br->num_BB_conflicts = 0;
    br->BB_conflict_list = NULL;

    inf_procs[bb->proc->id].inf_cfg[bb->id].branch = br;

    if( dbg & 0 ) { printBranch( br, 0 ); }
    return 0;
}
#endif

/*** MEMORY MODELING FRAMEWORK ***/

mem_p createMem(addr_t instAddr, reg_t writeAddr, reg_t regValue) {
    mem_p memBlk;
    memBlk = calloc(1,sizeof(mem_s));
    memBlk->instAddr = instAddr;
    initReg( &(memBlk->writeAddr) );
    initReg( &(memBlk->regValue) );
    cpyReg( &(memBlk->writeAddr),writeAddr);
    cpyReg( &(memBlk->regValue),regValue);
    return memBlk;
}

void freeMemBlk(mem_p *memBlk) {
    clrReg( &( (*memBlk)->writeAddr) ); 
    clrReg( &( (*memBlk)->regValue) ); 
    *memBlk=NULL;
}

void freeMemList(worklist_p *memList) {
    mem_p memBlk;
    worklist_p curNode, nextNode;
    curNode = *memList;
    while (curNode) {
        nextNode = curNode->next;
        memBlk = curNode->data;
        freeMemBlk( &memBlk );
        curNode->data = NULL;
        curNode->next = NULL;
        free(curNode);
        curNode = nextNode;
    }
    *memList = NULL;
}

int copyMem(mem_p dst, mem_p src) {
    worklist_p memNode;
    mem_p memBlk;
    int changed = 0;
    if (dst->instAddr != src->instAddr) {
        changed = 1;
        dst->instAddr = src->instAddr;
    }

    changed |= cpyReg( &(dst->writeAddr), src->writeAddr );
    changed |= cpyReg( &(dst->regValue), src->regValue );
    return changed;
}

void printMemList(FILE *fp,worklist_p memList) {
    worklist_p memNode;
    mem_p memBlk;
    fprintf(fp,"[ ");
    for (memNode = memList; memNode; memNode=memNode->next) {
        memBlk = memNode->data;
        if (memBlk->regValue.t==VALUE_UNPRED)continue;//dont print unpredictable
        fprintf(fp," (%lx ",memBlk->instAddr);
        printReg(fp,memBlk->writeAddr);
        fprintf(fp," ");
        printReg(fp,memBlk->regValue); fprintf(fp,")");
    }
    fprintf(fp," ]");
    fflush(fp);
}

/* dst = src, old dst data will be removed */
int copyMemList(worklist_p *dstList, worklist_p srcList) {
    int dbg = 0;
    worklist_p srcNode, dstNode, prvNode, memNode;
    mem_p srcBlk, dstBlk, memBlk;
    int changed,flag;
    srcNode = srcList; dstNode= *dstList; prvNode = NULL;
    changed = 0;
    if (dbg) {
        fprintf(dbgExec,"\n CopyMem: SRC ");
        printMemList(dbgExec,srcList);
    }
    while (srcNode) {
        if (dstNode) {
            flag = copyMem(dstNode->data, srcNode->data);
            if (flag) changed = 1;
            prvNode = dstNode;
            dstNode = dstNode->next;
        }
        else {
            dstBlk = calloc(1,sizeof(mem_s));
            copyMem(dstBlk, srcNode->data);
            addAfterNode(dstBlk, &prvNode, dstList);
            changed = 1;
        }
        srcNode = srcNode->next;
    }
    if (prvNode) {//remove remaining nodes of dst after prvNode
        prvNode->next = NULL;
        freeMemList( &dstNode);
        changed = 1;
    }
    else if (*dstList!=NULL) {//prvNode==NULL -> dstList=NULL
        freeMemList(dstList);
        changed = 1;
    }//dstList==NULL -> do nothing

    if (dbg) {
        fprintf(dbgExec," ->DST ");
        printMemList(dbgExec,*dstList);
    }
    return changed;
}

/*save deriValue to writeAddr to memList, for st instruction at addr instAddr*/
int storeMem (worklist_p *memList, addr_t instAddr,
                reg_t writeAddr, reg_t regValue) {
    int dbg=0;
    mem_p memBlk;
    worklist_p memNode, prvNode, addLoc;
    int updated, predictable, found;
    char str[DERI_LEN],des[DERI_LEN];
    FILE *dbgExec = stdout;

    //ERR: not handle unpredictable, assume it won't interfere with pred.mem 
    if (writeAddr.t==VALUE_UNPRED || writeAddr.t==VALUE_UNDEF) return 0;

#if 0
    if (regValue.t==VALUE_UNDEF || regValue.t==VALUE_UNPRED) predictable = 0;
    else predictable = 1;//type=CONST / EXPR / INDUCT
#endif
    /* cleekee: why not allow unpredictable value to be saved to mem? */
    predictable = 1;
    
    if (dbg && predictable) {
        fprintf(dbgExec,"\nStore: 0x%x ",instAddr);printReg(dbgExec, writeAddr);
        fprintf(dbgExec," ");printReg(dbgExec, regValue);
    }
    if (dbg && predictable) {
        fprintf(dbgExec,"\nBefore ");printMemList(dbgExec, *memList);
    }

    updated = 0;
    prvNode = NULL; addLoc = NULL;
    for (memNode = *memList; memNode; memNode=memNode->next) {
        memBlk = memNode->data;
        prvNode = memNode;
        if (regEq(memBlk->writeAddr, writeAddr)==1) break;
    }
    if (memNode) {//found memBlk->writeAddr == writeAddr
        if (!regEq(memBlk->regValue, regValue)) {
            if (memBlk->instAddr==instAddr) {
                //same inst, same writeAddr, diff. regValue -> merge value
                //mergeReg( &(memBlk->regValue), regValue, 0);
                cpyReg( &(memBlk->regValue), regValue);
                //regUnknown( &(memBlk->regValue) );
            }
            else {//set new regValue to memBlk->writeAddr
                memBlk->instAddr = instAddr;
                if (predictable) cpyReg( &(memBlk->regValue), regValue);
                else regUnknown( &(memBlk->regValue) );
            }
        }//else memBlk->regValue == regValue: do nothing
    }
    else {//cannot find writeAddr in memList
        if (predictable) {
            if (dbg) {
                fprintf(dbgExec,"\nAdd new mem entry");
                fprintf(dbgExec,"\nStore: 0x%x ",instAddr);
                printReg(dbgExec, writeAddr);
            }
            memBlk=createMem(instAddr,writeAddr,regValue);
            addAfterNode(memBlk,&prvNode,memList);
        }
    }
    if (dbg && predictable) {
        fprintf(dbgExec,"\nAfter: ");printMemList(dbgExec, *memList);}
    return updated;
}

/*load deriValue from writeAddr in memList, for ld instruction*/
boolean_t loadMem (worklist_p memList, reg_t writeAddr, reg_t *regValue) {
    int dbg=0;
    FILE *dbgExec = stdout;
    mem_p memBlk;
    worklist_p memNode;
    boolean_t memFound = 0;

    if (dbg) printMemList(dbgExec, memList);
    regUnknown(regValue);
    for (memNode=memList; memNode; memNode=memNode->next) {
        memBlk = memNode->data;
        if (regEq(writeAddr,memBlk->writeAddr)) {
            cpyReg(regValue, memBlk->regValue);
            //printf("Mem found: ");printReg(stdout, memBlk->regValue);printf(" type=%i\n",(memBlk->regValue).t);
            //printf("Mem changed: ");printReg(stdout, *regValue);printf(" type=%i\n",(*regValue).t);
            memFound = 1;
            break;
        }
    }

    /* cleekee: referenced data not explicitly stored by store inst, could be in code */
    if (!memFound) {
    	addr_t addr = writeAddr.val;
    	if (addr >= prog.code[0].addr && addr <= prog.code[prog.num_inst-1].addr) {
			regValue->t = VALUE_CONST;
			regValue->val = prog.code[(addr-prog.code[0].addr)/4].value;
			memFound = 1;
    	}
    }

    if (dbg && regValue->t!=VALUE_UNPRED) {
        fprintf(dbgExec,"\nLoad from i:%lx",memBlk->instAddr);
        fprintf(dbgExec," a:");printReg(dbgExec,memBlk->writeAddr);
        fprintf(dbgExec," v:");printReg(dbgExec,*regValue);
    }

    return memFound;
}

int memListEq(worklist_p dstList, worklist_p srcList) {
    int dbg = 0;
    worklist_p srcNode, dstNode;
    mem_p srcBlk, dstBlk;

    //not really correct, need to ensure certain order for the list
    //hopefully it can work, for now
    srcNode = srcList; dstNode = dstList;
    while (srcNode && dstNode) {
        srcBlk = srcNode->data; dstBlk=dstNode->data; 
        if (!regEq(srcBlk->writeAddr,dstBlk->writeAddr)
            || !regEq(srcBlk->regValue,dstBlk->regValue))
            return 0;
        srcNode = srcNode->next;
        dstNode = dstNode->next;
    }
    if (srcNode || dstNode) return 0;
    return 1;
}

int mergeMemList(worklist_p *dstList, worklist_p srcList) {
    int dbg = 0;
    worklist_p  memNode, srcNode, dstNode, srcPrv, dstTail, newList;
    mem_p   memBlk, srcBlk, dstBlk;
    int dstChanged;
    int cmpValue,flag;
    
    if (dbg) {
        fprintf(dbgExec,"\nMerge mem.list:");
        fprintf(dbgExec,"\nSrc: "); printMemList(dbgExec,srcList);
        fprintf(dbgExec,"\nDst: "); printMemList(dbgExec,*dstList);
    }
    dstChanged = 0;
    if (dbg && 0) fprintf(dbgExec,"\nMerging: ");
    dstTail = *dstList;
    while (dstTail && dstTail->next) dstTail = dstTail->next;
    newList = NULL;

    //a very simple but slow implementation of mem.search
    srcNode = srcList;
    while (srcNode) {
        srcBlk = srcNode->data;
        dstNode = *dstList;
        while (dstNode!=NULL) {
            dstBlk = dstNode->data;
            if (regEq(srcBlk->writeAddr,dstBlk->writeAddr)) {
                flag = mergeReg( &(dstBlk->regValue), srcBlk->regValue, 0);
                if (flag) dstChanged = 1;
                break;
            }
            dstNode = dstNode->next;
        }
        if (dstNode==NULL) {
            memBlk = createMem(srcBlk->instAddr,
                            srcBlk->writeAddr,srcBlk->regValue);
            addToWorkList(&newList,memBlk);
        }
        srcNode = srcNode->next;
    }
    if (dstTail) dstTail->next = newList;
    else *dstList = newList;

    if (dbg) {fprintf(dbgExec,"\nMERGED c:%d ",dstChanged);
                printMemList(dbgExec,*dstList);}
    return dstChanged;
}

// Note: for commutative operations on mixed-type operands
// (one is a variable and the other is a constant value),
// the derivation string is stored in the format "<var>+<const>"
#define PROCESS_STORE \
        if (traceMemFlag) {\
            /*HBK: sw -> save to mem model*/ \
            setInt(&tmpReg, atoi(insn->r2));\
            regOpr("+",&writeAddr,regList[r3],tmpReg);\
            storeMem(memList,instAddr,writeAddr,regList[r1]);\
            if (dbg) {\
                fprintf(dbgF,"\nStore (%x ",instAddr); \
                fprintf(dbgF," ");printReg(dbgF,writeAddr);\
                fprintf(dbgF," -> ");printReg(dbgF,regList[r1]);\
            }\
        }

#define PROCESS_LOAD \
        regUnknown(regList+r1);\
        if (traceMemFlag) {\
            /*HBK: ld -> load from mem model*/ \
            setInt(&tmpReg, atoi(insn->r2));\
            regOpr("+",&writeAddr,regList[r3],tmpReg);\
            loadMem(*memList,writeAddr,regList+r1);\
            if (dbg) {\
                fprintf(dbgF,"\nLoad (%x ",instAddr); \
                fprintf(dbgF," ");printReg(dbgF,writeAddr);\
                fprintf(dbgF," -> ");printReg(dbgF,regList[r1]);\
            }\
        }

#define PROCESS_SLTI \
        /* set r1 to 1 if r2 < atoi(r3) */ \
        if (regList[r2].t == VALUE_CONST) {\
            if (regList[r2].val<atoi(insn->r3)) \
                setInt(regList+r1,1);\
            else setInt(regList+r1,0);\
        }\
        else regUnknown(regList+r1);

#define PROCESS_SLT \
        /* set r1 to 1 if r2 < r3 */ \
        if (regList[r2].t == VALUE_CONST && regList[r2].t==regList[r3].t) {\
            if ( regList[r2].val < regList[r3].val ) \
                setInt(regList[r1],1);\
            else setInt(regList[r1],0);\
        }\
        else regUnknown(regList[r1]);

#define ADD_ASSIGN 
        /*add assignments for AB/BB conflict detection, ignore for now*/

int ignoreRegUpdate(char *r) {
    //$30 can be used as normal register when optimization flag is on
    if (streq(r,"$31")// || streq(r,"$30")|| streq(r,"$29")|| streq(r,"$28")
            || streq(r,"$0"))
        return 1;
    else return 0;
}

#ifdef _INF
static int regDerive(inf_node_t *ib, insn_t *insn, 
                int ignore, char *jal, int dbg) {
#else
static int regDerive(tcfg_node_t *bbi, insn_t *insn, 
                int ignore, char *jal, int dbg) {
#endif
    FILE        *dbgF = stdout;
    int         r1, r2, r3;
    addr_t      instAddr;
    insn_t      *insnTmp;
    biv_p       biv, bivInit; 
    char        *op;

    worklist_p  *memList;
    reg_t       *regList;
    reg_t       writeAddr, writeValue, tmpReg;  //for load/store

    initReg(&writeAddr);
    initReg(&writeValue);
    initReg(&tmpReg);

    ignore = 1; //no AB - BB detection supported, for now

#ifdef _INF
    memList = &(ib->memListOut);
    regList = (reg_t*)(ib->regListOut);
#else
	memList = &(bbi->memListOut);
	regList = (reg_t*)(bbi->regListOut);
#endif

    op = insn->op;
	
    sscanf(insn->addr,"%x",&instAddr);
    //if (dbg) printRegList(dbgF, (reg_t*)(ib->regListOut) );
    r1 = findReg(regList, insn->r1);
    r2 = findReg(regList, insn->r2);
    r3 = findReg(regList, insn->r3);
#if 0
    if( dbg ) { 
        fprintf(dbgF,"\n%s ",insn->r1);
        if (r1>=0) printReg(dbgF,regList[r1]); 
        fprintf(dbgF," %s ",insn->r2);
        if (r2>=0) printReg(dbgF,regList[r2]); 
        fprintf(dbgF," %s ",insn->r3);
        if (r3>=0) printReg(dbgF,regList[r3]); 
        fprintf(dbgF," | ");
    } 
#endif
#if 0
    if (deriveUnknown(regList,insn)) {
        if (dbg) { printf( "\n--> unpred. " ); printReg(stdout,regList[r1]);}
        return 1;
    }
#endif
    if (insn->flag==BIV_INST) {
        /*Derivation of inductive instruction*/
        if (dbg) {
			printf("\nInduction: ");
			printReg(stdout,regList[r1]);
		}

        if (regList[r1].t != VALUE_BIV) {
            biv = createBIV(regList, insn);
            if (regList[r1].t!=VALUE_UNDEF) {
                updateInitVal(biv,regList[r1]);
            }
            regList[r1].t = VALUE_BIV;
            regList[r1].biv = biv;
        } else {
            bivInit = regList[r1].biv;
            insnTmp = bivInit->insn;
            if (strcmp(insnTmp->addr,insn->addr) != 0) {
                //different biv of different instruction 
                biv = createBIV(regList, insn);
                updateInitVal(biv,regList[r1]);
                regList[r1].t = VALUE_BIV;
                regList[r1].biv = biv;
            }
            //else do nothing, induction value remains the same for loop
        }
		
        if (dbg) { 
			printf( "\n--> biv: " ); 
			printReg(stdout,regList[r1]);
		}
		
        return 1;
    }
#ifdef ARCH_ARM
    if (streq(op,"b")) {
        if (ignore) return 0;
        //add branches for AB / BB detection
    }
    else if (streq(op,"str") || streq(op,"strb")) {
		/* cleekee: currently does NOT handle post-indexed addressing */
        if (traceMemFlag) {
            setInt(&tmpReg, atoi(insn->r3));

            if (strcmp(regList[r2].name, "pc") == 0)
            	setInt(&regList[r2], (int)strtol(insn->addr, NULL, 16) + 8);

            regOpr("+",&writeAddr,regList[r2],tmpReg);
            storeMem(memList,instAddr,writeAddr,regList[r1]);
            if (dbg) {
                fprintf(dbgF,"\nStore (%x ",instAddr);
                fprintf(dbgF," ");printReg(dbgF,writeAddr);
                fprintf(dbgF," -> ");printReg(dbgF,regList[r1]);
            }
        }
        //ADD_ASSIGN;  
    }
    else if (ignoreRegUpdate(insn->r1)==1) {
        return 0;
    }
    else if (streq(op,"ldr") || streq(op,"ldrb")) {
		/* cleekee: currently does NOT handle post-indexed addressing */
        //regUnknown(regList+r1);
        if (traceMemFlag) {
            setInt(&tmpReg, atoi(insn->r3));

            if (strcmp(regList[r2].name, "pc") == 0)
            	setInt(&regList[r2], (int)strtol(insn->addr, NULL, 16) + 8);

            /* cleekee: some bug (likely in exploration) mess up "fp". Give default value if none. */
            if ((strcmp(regList[r2].name, "fp") == 0) && (regList[r2].t != VALUE_CONST))
            	setInt(&regList[r2], sp_init_value+4);

            regOpr("+",&writeAddr,regList[r2],tmpReg);

            boolean_t is_mem_loaded = loadMem(*memList,writeAddr,regList+r1);
            /* cleekee: if no mem found, create one and load to mem */
            if (!is_mem_loaded) {
            	tmpReg.t = VALUE_UNPRED;
            	storeMem(memList,instAddr,writeAddr,tmpReg);
            	loadMem(*memList,writeAddr,regList+r1);
            }

            if (dbg) {
                fprintf(dbgF,"\nLoad (%x ",instAddr);
                fprintf(dbgF," ");printReg(dbgF,writeAddr);
                fprintf(dbgF," -> ");printReg(dbgF,regList[r1]);
            }
        }
        //ADD_ASSIGN;  
    }
    else if(streq(op, "add")) {
		if (r3 == -1) {
			setInt(&tmpReg, atoi(insn->r3));

            if ((strcmp(regList[r2].name, "sp") == 0) && (regList[r2].t != VALUE_CONST))
            	setInt(&regList[r2], sp_init_value);

			regOpr("+",regList+r1,regList[r2],tmpReg);
		} else {
			regOpr("+",regList+r1,regList[r2],regList[r3]);
		}
        //ADD_ASSIGN;  
    }
    else if( streq(op, "sub") )
    {
		if (r3 == -1) {
			setInt(&tmpReg, atoi(insn->r3));
			regOpr("-",regList+r1,regList[r2],tmpReg);
		} else {
			regOpr("-",regList+r1,regList[r2],regList[r3]);
		}
        //ADD_ASSIGN;  
    }
    else if( streq(op, "lsl") )
    {
        setInt(&tmpReg,hexValue(insn->r3));
        regOpr("<<",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;
    }
    //TODO: cleekee: asl not the same as lsl?
    else if( streq(op, "asl") )
    {
        setInt(&tmpReg,hexValue(insn->r3));
        regOpr("<<",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;
    }
    else if( streq(op, "lsr") )
    {
        setInt(&tmpReg,hexValue(insn->r3));
        regOpr(">>",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;
    }
    //TODO: cleekee: asr not the same as lsr?
    else if( streq(op, "asr") )
    {
        setInt(&tmpReg,hexValue(insn->r3));
        regOpr(">>",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;
    }
	else if (streq(op,"mov")) {
        setInt(regList+r1, atoi(insn->r2));
        //ADD_ASSIGN;
    }
	else if (streq(op,"mvn")) {
        setInt(regList+r1, -(atoi(insn->r2)+1));
        //ADD_ASSIGN;
    }
	/* cleekee: not handle some other opcodes at the moment */
#if 0
    else if (streq(op,"ori")) {
        setInt(&tmpReg, atoi(insn->r3));
        regOpr("|",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;
    }
    else if( streq(op, "xori") )
    {
        setInt(&tmpReg, atoi(insn->r3));
        regOpr("^",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "andi") )
    {
        setInt(&tmpReg,atoi(insn->r3));
        regOpr("&",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "sll") )
    {
        setInt(&tmpReg,hexValue(insn->r3));
        regOpr("<<",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "srl") || streq(op,"sra") )
    {
        setInt(&tmpReg,hexValue(insn->r3));
        regOpr(">>",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "slti") || streq(op, "sltiu") )
    {
        PROCESS_SLTI;
        //ADD_ASSIGN;  
    }
    else if( streq(op, "slt") || streq(op, "sltu") )
    {
        //PROCESS_SLT;
        ADD_ASSIGN;  
    }
    else if( streq(op, "or") )
    {
        regOpr("^",regList+r1,regList[r2],regList[r3]);
        //ADD_ASSIGN;  
    }
#endif
    else {
        if (dbg) printf( "Ignoring opcode: %sn", op );
		//cleekee: temporary condition checking
        if (r1 >= 0) 
			regUnknown(regList+r1);
    }
#else
    if (streq(op,"beq") || streq(op,"bne")) {
        if (ignore) return 0;
        //add branches for AB / BB detection
    }
    else if (streq(op,"bgez") || streq(op,"blez") || streq(op,"bgtz")) {
        if (ignore) return 0;
        //add branches for AB / BB detection
    }
    else if ( streq(op,"sw")||streq(op,"sb")||streq(op,"sh")||streq(op,"sbu") ){
        PROCESS_STORE;
        //ADD_ASSIGN;
    }
    else if (ignoreRegUpdate(insn->r1)==1) {
        return 0;
    }
    else if (streq(op,"lw")||streq(op,"lb")||streq(op,"lbu")
            ||streq(op,"l.d")||streq(op,"lh")) {
        PROCESS_LOAD;
        //ADD_ASSIGN;
    }
    else if (streq(op,"lui")) {
        setInt(regList+r1, atoi(insn->r2)*65536);
        //ADD_ASSIGN;
    }
    else if( streq(op, "addiu") || streq(op, "addi") )
    {
        setInt(&tmpReg, atoi(insn->r3));
        regOpr("+",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;  
    }
    else if (streq(op,"ori")) {
        setInt(&tmpReg, atoi(insn->r3));
        regOpr("|",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;
    }
    else if( streq(op, "xori") )
    {
        setInt(&tmpReg, atoi(insn->r3));
        regOpr("^",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "andi") )
    {
        setInt(&tmpReg,atoi(insn->r3));
        regOpr("&",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "sll") )
    {
        setInt(&tmpReg,hexValue(insn->r3));
        regOpr("<<",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "srl") || streq(op,"sra") )
    {
        setInt(&tmpReg,hexValue(insn->r3));
        regOpr(">>",regList+r1,regList[r2],tmpReg);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "slti") || streq(op, "sltiu") )
    {
        PROCESS_SLTI;
        //ADD_ASSIGN;  
    }
    else if( streq(op, "slt") || streq(op, "sltu") )
    {
        //PROCESS_SLT;
        ADD_ASSIGN;  
    }
    else if( streq(op, "addu") )
    {
        regOpr("+",regList+r1,regList[r2],regList[r3]);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "subu") )
    {
        regOpr("-",regList+r1,regList[r2],regList[r3]);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "or") )
    {
        regOpr("^",regList+r1,regList[r2],regList[r3]);
        //ADD_ASSIGN;  
    }
    else if( streq(op, "jal") ) *jal = 1;
    else if( streq(op, "j") );
    else {
        if (dbg) printf( "Ignoring opcode: %sn", op );
		//cleekee: temporary condition checking
        if (r1 >= 0) 
			regUnknown(regList+r1);
    }
#endif
    if (dbg) {
    	if (r1 >= 0) {
			fprintf(dbgF,"\n ==> %s ",regList[r1].name);
			printReg(dbgF,regList[r1]);
    	}
    }

    clrReg(&tmpReg);
    clrReg(&writeAddr);
    clrReg(&writeValue);
    return 1;
}

/* join different incoming mem/reg values into one representative value
 * For loop header (lpHeader = 1), we ignore incoming values outside the loop, 
 * only join values from the back-edge. By merging values 
 * from current edge & back edge, we merge values from 
 * different iterations into one and identify loop invariants.
 * Non-invariant becomes unpredictable*/

#ifdef _INF
static int joinBranch(inf_proc_t *iproc, inf_node_t *ib, int isLpHead) {
    int dbg = 0;
    int active_reg = 14;
    int i,j,id;

    cfg_node_t *bb, *src;
    inf_node_t *isrc;
    cfg_edge_t *edge;
    reg_t *regSrc, *regCur, *regList;

    int bflag, flag, cmpFlag;
    biv_p biVar; 
    loop_t *lp;
    int cpFwd;

    char str[12];

    bb = ib->bb;
    lp = getIbLoop(ib);

		if (dbg) {
        printf("\nMerge incoming of c%d.%d, e:%d sa:%s, lpHeader %d ",
                iproc->proc->id,bb->id,cntExec,ib->insnlist[0].addr,isLpHead);
        fprintf(dbgExec,"\nMerge incoming of c%d.%d, e:%d sa:%s lpHeader %d",
                iproc->proc->id,bb->id,cntExec,ib->insnlist[0].addr,isLpHead);
    }
    
    if (dbg && 0) {
        printf("\nBefore merge\n");
        regList = (reg_t*)(ib->regListIn);
        for (j=11; j<26; j++) {printf("Reg %d: ",j);printReg(stdout,regList[j]);}
    }

    bflag = 0;
    
		/*merge all out. values of predecessor to in. value of current blk*/
    for (i=0; i<bb->num_in; i++) {
        edge = bb->in[i];
        src = edge->src;
        isrc = &(iproc->inf_cfg[src->id]);
        if (isLpHead) {//for lpHeader node: ignore in values from outside lp
            if (src->id < bb->id) continue; 
        }
        if (dbg) {
            fprintf(dbgExec,"\nsrc c%d.%d L:%d, dst: c%d.%d L:%d",
                iproc->proc->id, src->id, isrc->loop_id, 
                iproc->proc->id, bb->id, ib->loop_id);
        }

        regCur = (reg_t*)(ib->regListIn);
        regSrc = (reg_t*)(isrc->regListOut);
        /*merge memory value*/
        flag = mergeMemList( &(ib->memListIn), isrc->memListOut);
        if (dbg && flag) fprintf(dbgExec,"\n Changed in mem value");
        if (flag) bflag = 1;
        /*merge register value*/
        if (dbg) fprintf(dbgExec,"\nMerge register list");
        for (j=0; j< NO_REG; j++) {
            if (dbg && !regEq(regCur[j],regSrc[j])) {
                fprintf(dbgExec,"\n %s c%d.%d ",regCur[j].name,
                            iproc->proc->id,bb->id);
                printReg(dbgExec,regCur[j]);
                fprintf(dbgExec," <= rS: c%d.%d ",iproc->proc->id,src->id);
                printReg(dbgExec,regSrc[j]);
            }
            flag = mergeReg( regCur+j, regSrc[j], isLpHead );
            if (dbg && flag) {
                fprintf(dbgExec," => %s = ",regCur[j].name);
                printReg(dbgExec,regCur[j]);
            }
            if (flag) bflag = 1;
        }

    }

    if (dbg && 0) {
        printf("\nAfter merge, change:%d\n",bflag);
        regList = (reg_t*)(ib->regListIn);
        for (j=11; j<26; j++) {printf("Reg %d:",j);printReg(stdout,regList[j]);}
    }
    return bflag;
}
#else
static int joinBranch(tcfg_node_t* bbi, int isLpHead) {
    int dbg = 0;
    int active_reg = 14; /* sudiptac: what the hell is this 14? */
    int i,j,id;

    //cfg_node_t *bb;
		tcfg_node_t *src;
    inf_node_t *isrc;
    tcfg_edge_t *edge;
    reg_t *regSrc, *regCur, *regList;

    int bflag, flag, cmpFlag;
    biv_p biVar; 
    loop_t *lp;
    int cpFwd;

    char str[12];

		
    /* bb = ib->bb;
    lp = getIbLoop(ib); */

		lp = loop_map[bbi->id];

		bflag = 0;
    
		/*merge all out. values of predecessor to in. value of current blk*/
    for (edge = bbi->in; edge; edge = edge->next_in) {
        src = edge->src;
				/* sudiptac: should no longer be needed as we are shifting to
				tcfg structure to have register information */
        /* isrc = &(iproc->inf_cfg[src->id]); */
        if (isLpHead) {/* for lpHeader node: ignore in values from outside lp */
            //if (src->id < bbi->id) continue;
        	if (loop_map[src->id] != loop_map[bbi->id]) continue;
        }
        regCur = (reg_t*)(bbi->regListIn);
        regSrc = (reg_t*)(src->regListOut);
        /*merge memory value*/
        flag = mergeMemList( &(bbi->memListIn), src->memListOut);
        if (flag) bflag = 1;
        /*merge register value*/
        for (j = 0; j < NO_REG; j++) {
            flag = mergeReg( regCur+j, regSrc[j], isLpHead );
            if (flag) 
							bflag = 1;
        }
    }

		return bflag;
}
#endif

/*Check if insn is a base loop induction insn $1 = $1 + k*/
extern int gcd(int a, int b);

/*** Check basic induction format ***
 * i = i + k 
 * addu $1 $1 $2, $2=k
 * addi $1 $1 k
 * return k
 * return INFINITY if k not defined
 ***/
/*Derivation of unknown value, in addition to available derive*/
int deriveUnknown(reg_t *regList, insn_t *insn) {
    int dbg = 0;
    int flag;

    int regPos1, regPos2, regPos3;
    regPos1 = findReg(regList, insn->r1);
    regPos2 = findReg(regList, insn->r2);
    regPos3 = findReg(regList, insn->r3);
    flag = 0;
    if ( (regPos2>-1 && regList[regPos2].t==VALUE_UNPRED)
        && (regPos3>-1 && regList[regPos3].t==VALUE_UNPRED) ) {
        regUnknown(regList+regPos1);
        flag = 1;
    }
    return flag;
}

int biv2LoopId(biv_p biv) {
	return loop_map[((insn_t*)(biv->insn))->bbi->id]->id;
#if 0
    int         dbg = 0;
    int         i, j;
    int         bivAddr, lpAddrStr, lpAddrEnd;
    cfg_node_t  *bb;
    insn_t      *insn;
    loop_t      *lp;

    insn = (insn_t*)(biv->insn);
    for (i=1; i<num_tcfg_loops; i++) {
        lp = loops[i]; 

        bb = lp->head->bb;
        lpAddrStr = bb->sa;

        lpAddrEnd = lpAddrStr;
        for (j=0; j<lp->num_tails; j++) {
        	bb = lp->tails[j]->bb;
        	if (lpAddrEnd < bb->sa+bb->size-1)
        		lpAddrEnd = bb->sa+bb->size-1;
        }

        bivAddr = hexValue(insn->addr);
        if (lpAddrStr <= bivAddr && bivAddr<lpAddrEnd) {
            if (dbg) {
                fprintf(dbgExec,"\n BIV inst 0x%s in L%d ",insn->addr,lp->id);
                printBIV(dbgExec,biv);
            }
            return lp->id;
        }
    }

    return -1; //cannot find
#endif
}

/*Expand addrExpr to non-recursive, non-biv expression*/
void expandAddrExpr(expr_p exp, reg_t  *addrExpr) {
    int         dbg = 0;
    reg_t*      curReg;
    worklist_p  expandList;
    biv_p       biv;
    int         i, varNum, lpId;

    if (dbg) {fprintf(dbgExec,"\n----------\n");fprintf(dbgExec,"AddrExpr: ");printReg(dbgExec,*addrExpr);}
    clrExpr(exp);
    expandList = NULL;
    addToWorkList(&expandList,addrExpr);
    while (!isEmpty(expandList)) {
        curReg = (reg_t*) removeOneFromWorkList(&expandList);
        if (dbg){fprintf(dbgExec,"\nExpand reg: ");printReg(dbgExec,*curReg);}
        if (dbg) {fprintf(dbgExec,"\nCur expr: ");printExpr(dbgExec,exp);}
        switch (curReg->t) {
            case VALUE_UNDEF:
                fprintf(dbgExec,"\nWarning: unknown parameter in addrExpr");
                varNum = exp->varNum;
                strcpy( exp->value[varNum].para, "T" );
                exp->coef[varNum] = 1;
                exp->varNum++;
                break;
            case VALUE_CONST:
            	if (dbg) fprintf(dbgExec,"\nExpanded to constant");
                exp->k += curReg->val;
                break;
            case VALUE_PARA:
                fprintf(dbgExec,"\nWarning: unknown parameter in addrExpr");
                varNum = exp->varNum;
                cpyReg( &(exp->value[varNum]), *curReg);
                exp->coef[varNum] = 1;
                exp->varNum++;
                break;
            case VALUE_BIV:
            	if (dbg) fprintf(dbgExec,"\nExpanded to BIV");
                //biv is represented in addrExpr as lpId * stride
                varNum = exp->varNum;
                lpId = biv2LoopId(curReg->biv);
                setInt( &(exp->value[varNum]), lpId );
                //cpyReg( &(exp->value[varNum]), *curReg);
                exp->coef[varNum] = curReg->biv->stride;
                if (dbg) fprintf(dbgExec," %d:L%d",
                                exp->coef[varNum],exp->value[varNum].val);
                exp->varNum++;
                addToWorkList(&expandList,&(curReg->biv->initVal));
                break;
            case VALUE_EXPR:
            	if (dbg) fprintf(dbgExec,"\nExpanded to expr");
                exp->k += curReg->expr->k;
                for (i=0; i<curReg->expr->varNum; i++) {
                    varNum = exp->varNum;
                    exp->coef[varNum] = curReg->expr->coef[i];
                    if (curReg->expr->value[i].t==VALUE_BIV) {
                        biv = curReg->expr->value[i].biv;
                        lpId = biv2LoopId(biv);
                        setInt( &(exp->value[varNum]), lpId );
                        exp->coef[varNum] *= biv->stride;
                        if (dbg) fprintf(dbgExec," %d:L%d",
                                exp->coef[varNum],exp->value[varNum].val);
                        addToWorkList(&expandList,&(biv->initVal));
                    }
                    else {
                        cpyReg( &(exp->value[varNum]), curReg->expr->value[i]); 
                    }
                    exp->varNum++;
                }
                break;
            default:
                printf("\n Panic (expandBIV) , unknown type %d",curReg->t);
                exit(1);
        }
        if (exp->varNum==MAX_EXPR_LEN) {
            printf("\nPanic, exceed MAX_EXPR_LEN ");
            exit(1);
        }
    }
    if (dbg) {fprintf(dbgExec,"\nExpanded expr: ");printExpr(dbgExec,exp);fprintf(dbgExec,"\n----------\n");}
}
#ifdef _INF
static void readSymExecResult(inf_node_t *ib) {
    int dbg = 0;
    int k, r1,r2,r3;
    insn_t  *insn;
    reg_t   *listIn, *listOut;
    char    jal, ignore;  // procedure call cancels conflict
    reg_t   tmpReg, exprReg;
    initReg(&tmpReg);initReg(&exprReg);

    cfg_node_t  *bb;
    dat_inst_t  *dat;
    int         datNum;

    /*init data inst in bb node*/
    bb = ib->bb;
    if (ib->num_insn!=bb->num_inst) {printf("\nPanic ib!=bb"); exit(0);}

    bb->num_d_inst=0;
    for (k=0; k<ib->num_insn; k++) {
        insn = &(ib->insnlist[k]);
        if (isLoadInst(insn->op) || isStoreInst(insn->op)) bb->num_d_inst++;
    }
    bb->d_instlist = calloc(bb->num_d_inst,sizeof(dat_inst_t));

    //pairDetectionFlag = 1;

    listIn = (reg_t*)(ib->regListIn);
    listOut = (reg_t*)(ib->regListOut);
    copyMemList( &(ib->memListOut), ib->memListIn);
    for ( k = 0; k < NO_REG; k++ ) {
        cpyReg(listOut+k, listIn[k]);
    }

    datNum = 0;
    for ( k = 0; k < ib->num_insn; k++ ) {
        insn = &(ib->insnlist[k]);
        if (isLoadInst(insn->op) || isStoreInst(insn->op)) {
            r1 = findReg(listOut, insn->r1);
            r2 = findReg(listOut, insn->r2);
            r3 = findReg(listOut, insn->r3);
            if (dbg) {
                printf("\n ");printInstr(insn);
                #if 0
                printf("\n%s ",insn->r1);
                if (r1>=0) printReg(stdout,listOut[r1]); 
                printf(" %s ",insn->r2);
                if (r2>=0) printReg(stdout,listOut[r2]); 
                #endif
                printf("| %s = ",insn->r3);
                if (r3>=0) printReg(stdout,listOut[r3]); 
                printf("\n");
            }
            setInt(&tmpReg, atoi(insn->r2));
					
						/* sudiptac: intra-procedural limitations */
            dat             = &(((dat_inst_t*)(bb->d_instlist))[datNum]);
            datNum++;
            dat->insn       = &(ib->insnlist[k]);
            dat->addr_set   = NULL;
            dat->resideLpId = getIbLoop(ib)->id;
            regOpr("+",&exprReg,listOut[r3],tmpReg);
						/* sudiptac: intra-procedural limitation */
            initExpr( &(dat->addrExpr) );
            expandAddrExpr( &(dat->addrExpr), &exprReg);
            if (dbg) {printDataRef(dbgExec,dat);}
        }
        ignore = jal; jal = 0;
        regDerive(ib,insn,ignore,&jal,0);/*derive normally*/
    }
    clrReg(&tmpReg);clrReg(&exprReg);
}
#else
static void readSymExecResult(tcfg_node_t *bbi) {
    int dbg = 0;
    int k, r1,r2,r3;
    insn_t  *insn;
    reg_t   *listIn, *listOut;
    char    jal, ignore;  // procedure call cancels conflict
    reg_t   tmpReg, exprReg;
	
    initReg(&tmpReg);
	initReg(&exprReg);

    cfg_node_t  *bb;
    dat_inst_t  *dat;
    int         datNum;

    /*init data inst in bb node*/
    bb = bbi->bb;
    bb->num_d_inst=0;
    
	for (k = 0; k < bb->num_inst; k++) {
        insn = &(bbi->insnlist[k]);
        if (isLoadInst(insn->op) || isStoreInst(insn->op)) 
			bb->num_d_inst++;
    }
    bbi->d_instlist = calloc(bbi->bb->num_d_inst,sizeof(dat_inst_t));

    //pairDetectionFlag = 1;

    listIn = (reg_t*)(bbi->regListIn);
    listOut = (reg_t*)(bbi->regListOut);
    copyMemList( &(bbi->memListOut), bbi->memListIn);
    
	for ( k = 0; k < NO_REG; k++ ) {
        cpyReg(listOut+k, listIn[k]);
    }

    datNum = 0;
    
	for ( k = 0; k < bb->num_inst; k++ ) {
        insn = &(bbi->insnlist[k]);

        /////
        if (dbg) {
			r1 = findReg(listOut, insn->r1);
			r2 = findReg(listOut, insn->r2);
			r3 = findReg(listOut, insn->r3);
			printf("\n ");
			printInstr(insn);
			printf(" (flag=%i) ",insn->flag);
			printf(" | %s = ",insn->r2);
			if (r2>=0) printReg(stdout,listOut[r2]);
			printf("\n");

			printf("arg2: %s (type=%i) ", insn->r2, listOut[r2].t); printReg(stdout,listOut[r2]); printf("\n");
			printf("arg3: %s (type=%i) ", insn->r3, listOut[r3].t); printReg(stdout,listOut[r3]); printf("\n");
        }
        /////

        if (isLoadInst(insn->op) || isStoreInst(insn->op)) {
            r1 = findReg(listOut, insn->r1);
            r2 = findReg(listOut, insn->r2);
            r3 = findReg(listOut, insn->r3);
            if (dbg) {
                printf("\n ");
				printInstr(insn);
                printf(" | %s = ",insn->r2);
                if (r2>=0) printReg(stdout,listOut[r2]); 
                printf("\n");
            }
#ifdef ARCH_ARM
            setInt(&tmpReg, atoi(insn->r3));

            if (strcmp(listOut[r2].name, "pc") == 0)
            	setInt(&listOut[r2], (int)strtol(insn->addr, NULL, 16) + 8);
#else
            setInt(&tmpReg, atoi(insn->r2));
#endif	
			/* sudiptac: intra-procedural limitations */
            dat             = &(((dat_inst_t*)(bbi->d_instlist))[datNum]);
            datNum++;
            dat->insn       = &(bbi->insnlist[k]);
            dat->addr_set   = NULL;
            dat->resideLpId = loop_map[bbi->id]->id;
#ifdef ARCH_ARM
            regOpr("+",&exprReg,listOut[r2],tmpReg);
#else
            regOpr("+",&exprReg,listOut[r3],tmpReg);
#endif	
			/* sudiptac: intra-procedural limitation */
            initExpr( &(dat->addrExpr) );
            expandAddrExpr( &(dat->addrExpr), &exprReg);
            if (dbg) {printDataRef(dbgExec,dat);}
        }
        ignore = jal; jal = 0;
        regDerive(bbi,insn,ignore,&jal,0);/*derive normally*/

        /////
        if (dbg) {
        	printf("arg1: %s (type=%i) ", insn->r1, listOut[r1].t); printReg(stdout,listOut[r1]); printf("\n");
        }
        /////
    }
    clrReg(&tmpReg);clrReg(&exprReg);
}
#endif

#ifdef _INF
static void analyzeBlock(inf_node_t *ib) {
    int     dbg = 0;
    int     i, j, k, m, id;
    int     r1, regPos2, regPos3;


    char    jal, ignore;  // procedure call cancels conflict
    char    flag;
    insn_t  *insn;
    reg_t   *regList, *listIn, *listOut, tmpReg;
    FILE    *fp = stdout;

    jal = 1;

    initReg(&tmpReg);
    if (dbg) {
        fprintf(dbgExec,"\nAnalyze block (%d,%d)", ib->bb->proc->id, ib->bb->id);
        printf("\nAnalyze block (%d,%d)", ib->bb->proc->id, ib->bb->id);
        if (0) printMemList(fp,ib->memListOut);
        fflush(dbgExec);
    }

    //copy in-status to out-status
    copyMemList(&(ib->memListOut), ib->memListIn);
    listIn = (reg_t*)(ib->regListIn);
    listOut= (reg_t*)(ib->regListOut);
    for (j=0;j<NO_REG;j++) cpyReg(listOut+j,listIn[j]);
    regList = (reg_t*)(ib->regListOut);
    for( k = 0; k < ib->num_insn; k++ ) {
        insn = &(ib->insnlist[k]);
#if 0
        if(dbg) { 
            //printf( "\n[%2d,%2d,%2d] ", i, j, k ); 
            printf("\n k:%d ",k); printInstr( insn ); 
            //fprintf(dbgExec,"\n k:%d ",k);fprintInstr( dbgExec, insn ); 
        }
#endif
        ignore = jal; jal = 0;
        regDerive(ib,insn,ignore,&jal,dbg);/*derive normally*/
#if 0
        if (dbg) printf("\n");
#endif
    } // end for instr
#if 0
    if (dbg) {
        fprintf(dbgExec,"\nFinish block (%d,%d)", ib->bb->proc->id, ib->bb->id);
        printf("\nFinish block (%d,%d)", ib->bb->proc->id, ib->bb->id);
        if (0) printMemList(fp,ib->memListOut);
        fflush(dbgExec);fflush(stdout);
    }
#endif
}
#else
static void analyzeBlock(tcfg_node_t *bbi) {
    int     dbg = 0;
    int     i, j, k, m, id;
    int     r1, regPos2, regPos3;

    char    jal, ignore;  // procedure call cancels conflict
    char    flag;
    insn_t  *insn;
    reg_t   *regList, *listIn, *listOut, tmpReg;
    FILE    *fp = stdout;

    jal = 1;

    initReg(&tmpReg);
    
	if (dbg) {
        fprintf(dbgExec,"\nAnalyze block (%d,%d)", bbi->bb->proc->id, bbi->bb->id);
        printf("\nAnalyze block (%d,%d)", bbi->bb->proc->id, bbi->bb->id);
        if (0) printMemList(fp,bbi->memListOut);
        fflush(dbgExec);
    }

    //copy in-status to out-status
    copyMemList(&(bbi->memListOut), bbi->memListIn);
    listIn = (reg_t*)(bbi->regListIn);
    listOut= (reg_t*)(bbi->regListOut);
    
	for (j = 0; j < NO_REG; j++)
		cpyReg(listOut+j, listIn[j]);
    
	regList = (reg_t*)(bbi->regListOut);

	for( k = 0; k < bbi->bb->num_inst; k++ ) {
		/* sudiptac: move "insnlist" to tcfg_node structure */
        insn = &(bbi->insnlist[k]);
        ignore = jal; jal = 0;

        regDerive(bbi,insn,ignore,&jal,dbg);/*derive normally*/

    } // end for instr
}
#endif

/*Check if regName is redefined in insn*/
static int regRedefined(char* regName, insn_t *insn) {
    char *op = insn->op;
    int  regPos1;

    /*regName is redefined if (insn r1=regName, insn is assigment*/
    if (strcmp(regName, insn->r1)!=0) return 0;
#ifdef ARCH_ARM
    if (streq(op,"ldr") || streq(op,"add") || streq(op,"sub") || streq(op,"mov") )
        return 1;
#else
    if (streq(op,"lw") || streq(op,"lb") || streq(op,"lh") || streq(op,"lbu"))
        return 1;
    if (streq(op,"addu") || streq(op,"addiu") || streq(op,"addi"))
        return 1;
    if (streq(op,"ori") || streq(op,"xori") || streq(op,"andi"))
        return 1;
    if (streq(op,"sll") || streq(op,"srl") || streq(op,"slti") 
            || streq(op,"sltu")) return 1;
    if (streq(op,"subu") || streq(op,"or") || streq(op,"xor"))
        return 1;
#endif
    return 0;
}

#ifdef _INF
static void reInit(inf_proc_t* p, int startId, int endId, int *visited) {
    int dbg = 0;
    inf_node_t *ib;
    int i;
    //reinitialize everything within the loop
    for (i=startId; i<=endId; i++) {
        visited[i]=0;
        if (i==0) continue;//bb0 has no incoming block
        ib = &(p->inf_cfg[i]); 
        clearRegList(ib->regListIn);
        clearRegList(ib->regListOut);
        freeMemList( &(ib->memListIn));
        //ib->memListIn = NULL; //ERR: memory leak
        freeMemList( &(ib->memListOut));
        //ib->memListOut = NULL; //ERR: memory leak
        if (dbg) {
            fprintf(dbgExec,"\nReinit c%d.%d",p->proc->id,ib->bb->id);
            printMemList(dbgExec,ib->memListIn);
            printMemList(dbgExec,ib->memListOut);
            fflush(dbgExec);
        }
    }
}
#else
static void reInit(int startId, int endId, int *visited, int headExcluded) {
    int dbg = 0;
    tcfg_node_t *bbi;
    int i;
    
    //for (i = startId; i <= endId; i++) {
    for (i=0; i<num_tcfg_nodes; i++) {
    	if (loop_comm_ances[loop_map[i]->id][loop_map[startId]->id] != loop_map[startId])
    		continue;

    	if (headExcluded && startId==loop_map[startId]->head->id)
    		continue;

        visited[i]=0;
        if (i==0) 
					continue; /* bb0 has no incoming block */
        bbi = tcfg[i]; 
        clearRegList(bbi->regListIn);
        clearRegList(bbi->regListOut);
        freeMemList(&(bbi->memListIn));
        //ib->memListIn = NULL; //ERR: memory leak
        freeMemList(&(bbi->memListOut));
        //ib->memListOut = NULL; //ERR: memory leak
        if (dbg) {
            fprintf(dbgExec,"\nReinit c%d.%d",bbi->bb->proc->id,bbi->bb->id);
            printMemList(dbgExec,bbi->memListIn);
            printMemList(dbgExec,bbi->memListOut);
            fflush(dbgExec);
        }
    }
}
#endif

/*** Register expansion framework to trace reg. value and induction var ***
 * Detect loop induction variable (w/o initial value)
 * Perform register expansion (except for induction var)
 * Collect init value of induction vars
 * Collect AB - BB conflicting pairs 
 **/
/*symbolic execution within a particular loop*/
#ifdef _INF
int loopSymExec(loop_t *lp, inf_proc_t *p, int mustFlag, P_Queue **outQueue) {
#else
int loopSymExec(loop_t* lp, tcfg_node_t * bbi, int mustFlag, P_Queue** outQueue) {
#endif
    int dbg = 0;
    int i,j,k;
    int lpTailId, lpHeadId;

#ifdef _INF
    inf_node_t *inNode;
#else
		tcfg_node_t* inNode;
#endif

    loop_t *inLp;
    inf_node_t *ib;
    cfg_node_t *src, *dst;
    int pchange, bbchange;
    reg_t * listIn, *listOut;

    P_Queue *pQueue,*qElem;
    int ibId, lpId;
    int *pInt;
    int visited[MAX_OVRL_NODES];
		tcfg_node_t* bbihead;
		tcfg_edge_t* edge;

#ifdef _INF
    if (lp) lpHeadId = lp->head->bb->id;
#else
		if (lp) lpHeadId = lp->head->id;
#endif
    else lpHeadId = 0; //outermost loop

#ifdef _INF
    if (lp && lp->tail) lpTailId = lp->tail->bb->id;
    else lpTailId = p->num_bb-1;
#else
    if (lp && lp->tail) lpTailId = lp->tail->id;
		else lpTailId = num_tcfg_nodes - 1;
#endif

    inNode = NULL;
    if (lp) lpId = lp->id; else lpId = 0;

    //check if fixed-point - no change in in-status
    pchange = 0;

#ifdef _INF
    ib = &(p->inf_cfg[lpHeadId]);
    listIn = (reg_t*)(ib->regListIn);
#else
		bbihead = tcfg[lpHeadId];
		listIn = (reg_t *)(bbihead->regListIn);
#endif
    
	if (dbg) {
		fprintf(dbgExec,"\n----- Sym.exec of loop L%d, [%d,%d], mustFlag=%d -----",
				lpId,lpHeadId,lpTailId,mustFlag);
		printf("\nSym.exec of loop L%d, [%d,%d], %d",
				lpId,lpHeadId,lpTailId,mustFlag);
	}

#ifdef _INF
		for (i=0; i<ib->bb->num_in; i++) {
        inNode = &(p->inf_cfg[ib->bb->in[i]->src->id]);
        listOut = (reg_t*)(inNode->regListOut);
        for (k=0; k<NO_REG; k++) 
            if ( !regEq(listIn[k],listOut[k]) ) {
                if (0) {
                    printf("\nRegIn ");
										printReg(stdout, listOut[k]);
                    printf("\nRegCur ");
										printReg(stdout,listIn[k]);
                }
                pchange = 1;
            }
    }
#else
		for (edge = bbihead->in; edge; edge = edge->next_in) {
				inNode = tcfg[edge->src->id];
				listOut = (reg_t *)(inNode->regListOut);
        for (k = 0; k < NO_REG; k++)
            if ( !regEq(listIn[k],listOut[k]) ) {
                if (0) {
                    fprintf(dbgExec, "\nRegIn ");
										printReg(dbgExec, listOut[k]);
                    fprintf(dbgExec, "\nRegCur ");
										printReg(dbgExec,listIn[k]);
                }
                pchange = 1;
                if(dbg) fprintf(dbgExec,"\npchange = 1, listIn[%i] != listOut[%i]", k , k);
                break;
            }
		}
#endif
    if (!mustFlag && !pchange && inNode!=NULL) {
        if (dbg) fprintf(dbgExec,"\nNo change in loop L%d",lpId);
        if (dbg) printf("\nNo change in loop L%d",lpId);
        return 0;
    }

#ifdef _INF
    reInit(p, lpHeadId, lpTailId, visited);
#else
	reInit(lpHeadId, lpTailId, visited, 0);
#endif

    //copy in-status to loop head
#ifdef _INF
    ib = &(p->inf_cfg[lpHeadId]);
    listIn = (reg_t*)(ib->regListIn);
    if (ib->bb->num_in>0) {
        inNode = &(p->inf_cfg[ib->bb->in[0]->src->id]);
        listOut = (reg_t*)(inNode->regListOut);
#else
    bbihead = tcfg[lpHeadId];
    listIn = (reg_t*)(bbihead->regListIn);
    if (bbihead->in) {
		inNode = tcfg[bbihead->in->src->id];
        listOut = (reg_t*)(inNode->regListOut);	
#endif

#ifdef _INF
        if (dbg) {
            printf("\nCopy ACS from loop pre-header  c%d.%d -> c%d.%d",
                    p->proc->id,inNode->bb->id, p->proc->id, ib->bb->id);
            //printf("\nACS pre-header c%d.%d ",p->proc->id, inNode->bb->id);
            //printRegList(stdout,listOut);
        }
        copyMemList( &(ib->memListIn), inNode->memListOut);
#else
		copyMemList( &(bbihead->memListIn), inNode->memListOut);
#endif

        for (k=0; k<NO_REG; k++) {
            if (!regEq( listIn[k], listOut[k] )) {
                if (dbg) {
                    printf("\nChanged: %s, cur: ",listIn[k].name);
                    printReg(stdout,listIn[k]);
                    printf("new: "); printReg(stdout,listOut[k]);
                }
                cpyReg(listIn+k, listOut[k]);
            }
        }
    } else {
		inNode = NULL;
    }

    /* perform fixed-point sym.exec within the loop */
    while (1) {
        /* merge lpHeader input with previous iteration value */
#ifdef _INF
        ib = &(p->inf_cfg[lpHeadId]);
        bbchange = joinBranch(p,ib,1);
#else
		bbihead = tcfg[lpHeadId];
		/* sudiptac: changed to handle tcfg nodes */
		bbchange = joinBranch(bbihead, 1);

		if (dbg) fprintf(dbgExec, "\n  L%d header joinBranch with backedge, bbchange=%d", lpId, bbchange);
#endif
#if 0
        if (dbg) {
            fprintf(dbgExec,"\n  Start next L%d iter at lpHeader c%d.%d",lpId,
            		bbihead->bb->proc->id, bbihead->bb->id);
            printf("\n  Start next L%d iter at lpHeader c%d.%d",lpId,
            		bbihead->bb->proc->id, bbihead->bb->id);
        }
#endif
		if (bbchange==0 && visited[lpHeadId]==1) {
			fprintf(dbgExec, "\n  Return from L%d due to (bbchange==0 && visited[lpHeadId]==1)", lpId);
			break;
		}

        visited[lpHeadId] = 1;
#ifdef _INF
        reInit(p, lpHeadId+1, lpTailId, visited);
#else
		reInit(lpHeadId /*+ 1*/, lpTailId, visited, 1);
#endif
        pQueue = NULL;

        p_enqueue(&pQueue,&lpHeadId,lpHeadId);
        if (dbg) fprintf(dbgExec, "\n    Enqueued bbi %d", lpHeadId);

        while (!p_queue_empty(&pQueue)) { 
            cntExec++; //count total node sym.exec performed
            if (cntExec>=3000) {
                printf("\n    Too many iterations, no terminate?\n");
                exit(1);
            }
            pInt = (int*)p_dequeue(&pQueue);
			ibId = *pInt;

			if (dbg) fprintf(dbgExec, "\n    Dequeued bbi %d", ibId);

#ifdef _INF
            ib = &(p->inf_cfg[ibId]);
#else
			bbihead = tcfg[ibId];
#endif
#if 0
            if (dbg) {//print outgoing nodes of current node
                fprintf(dbgExec,"\n    A_ bb %d:%d (",ibId,ib->bb->id);
                if (ib->bb->out_n) {
                    dst = ib->bb->out_n->dst;
                    fprintf(dbgExec,"%d,",dst->id);fflush(dbgExec);
                }
                if (ib->bb->out_t) {
                    dst = ib->bb->out_t->dst;
                    fprintf(dbgExec,"%d,",dst->id);fflush(dbgExec);
                }
                fprintf(dbgExec,")\n");fflush(dbgExec);
            }
#endif
            //if (dbg) fprintf(dbgExec,"\nIncoming b%d ",ib->bb->id);
#ifdef _INF
            for (i = 0; i < ib->bb->num_in; i++) {
                /*Assume: bb->in[i]->id > bb->id --> loop back edge -> loop head*/
                inNode = &(p->inf_cfg[ib->bb->in[i]->src->id]);
                //if (dbg) fprintf(dbgExec," b%d ",inNode->bb->id);
                if (inNode->bb->id >= ib->bb->id) ib->loop_role = LOOP_HEAD;
            }
            if (ib->bb->id > lpHeadId && ib->loop_role == LOOP_HEAD) {
                //ib is lpHead of an inner loop 
                inLp = getIbLoop(ib);
                loopSymExec(inLp,p,!visited[ib->bb->id], &pQueue);
                //could make a dead loop
#if 0
                if (dbg) {
                    fprintf(dbgExec,"\n    Back to L%d [%d,%d]",lpId,lpHeadId,lpTailId);
                    printf("\n    Back to L%d [%d,%d]",lpId,lpHeadId,lpTailId);
                }
#endif
                visited[ib->bb->id] = 1;
                continue;
            }
#else
#if 0
            /* cleekee: this part try to find out whether bbihead is a loop head? */
            for (edge = bbihead->in; edge; edge = edge->next_in) {
                /*Assume: bb->in[i]->id > bb->id --> loop back edge -> loop head*/
				inNode = tcfg[edge->src->id];
                //if (inNode->id >= bbihead->id)
					//bbihead->bb->loop_role = LOOP_HEAD;
            }
#endif
            //if (bbihead->id > lpHeadId && bbihead->bb->loop_role == LOOP_HEAD) {
            if (loop_comm_ances[lp->id][loop_map[bbihead->id]->id] == lp
            	&& loop_map[bbihead->id] != lp
            	&& loop_map[bbihead->id]->head == bbihead) {
                //bbihead is lpHead of an inner loop
#ifdef _INF
                inLp = getIbLoop(ib);
                loopSymExec(inLp,p,!visited[ib->bb->id], &pQueue);
                //could make a dead loop
#else
				inLp = loop_map[bbihead->id];

				if (dbg) fprintf(dbgExec, "\n    bbi %d is header of inner lp of L%d", bbihead->id, lpId);

				/* sudiptac: assign the tcfg node here */
				loopSymExec(inLp, NULL, !visited[bbihead->id], &pQueue);

				if (dbg) fprintf(dbgExec, "\n    Continue with L%d", lpId);
#endif

#if 0
                if (dbg) {
                    fprintf(dbgExec,"\n    Back to L%d [%d,%d]",lpId,lpHeadId,lpTailId);
                    printf("\n    Back to L%d [%d,%d]",lpId,lpHeadId,lpTailId);
                }
#endif
                visited[bbihead->id] = 1;
					continue;
            }
#endif

#ifdef _INF
            /*Merge in-status of all predecessor bb*/
            if (ib->bb->id==lpHeadId) {
                bbchange = 1;//???
            }
            else {
                bbchange = joinBranch(p,ib, 0);
                if (dbg) fprintf(dbgExec,"\n    Join in-status of c%d.%d",
                        p->proc->id, ib->bb->id);
            }
#else
            /*Merge in-status of all predecessor bb*/
			/* sudiptac: I have no idea what is happening here, keeping
			it to maintain consistency with the previous implementation */
            if (bbihead->id==lpHeadId) {
                bbchange = 1; 
            }
            else {
				/* sudiptac: changed to handle tcfg nodes */
				bbchange = joinBranch(bbihead, 0);
                if (dbg) fprintf(dbgExec,"\n    Join in-status of c%d.%d",
                       	bbihead->bb->proc->id, bbihead->bb->id);
            }
#endif
#ifdef _INF
      			if (bbchange || visited[ib->bb->id]==0) {
                visited[ib->bb->id]=1;
                analyzeBlock(ib);
                //do not add loop header back to the queue
                if (ib->bb->out_n) {
                    dst = ib->bb->out_n->dst;
										//cleekee: tentative to change
										if (loop_map[dst->id] == lp)
                    		//if (lpHeadId < dst->id && dst->id <= lpTailId)
                        //within the current loop
                        p_enqueue(&pQueue,&(dst->id),dst->id);
										else if (dst->id > lpTailId)
                        //outside the current loop
                        p_enqueue(outQueue,&(dst->id),dst->id);
                    //fprintf(dbgExec,"Out_n %d\n",dst->id);fflush(dbgExec);
                }
                if (ib->bb->out_t) {
                    dst = ib->bb->out_t->dst;
								if (loop_map[dst->id] == lp)
                    //if (lpHeadId < dst->id && dst->id <= lpTailId)
                    //within the current loop
                    p_enqueue(&pQueue,&(dst->id),dst->id);
									else if (dst->id > lpTailId)
                    //outside the current loop
                    p_enqueue(outQueue,&(dst->id),dst->id);
                    //fprintf(dbgExec,"Out_t %d \n",dst->id);fflush(dbgExec);
                }
            }
#else
			if (bbchange || visited[bbihead->id]==0) {
				tcfg_node_t* dst;
				tcfg_edge_t* edge;

				visited[bbihead->id] = 1;

				/* sudiptac: changed to handle tcfg nodes instead of inf_ nodes */
				analyzeBlock(bbihead);

				if (dbg) fprintf(dbgExec, "\n    Analyzed bbi %d", bbihead->id);

				for (edge = bbihead->out; edge; edge = edge->next_out) {
					dst = edge->dst;

					//if (lpHeadId < dst->id && dst->id <= lpTailId) {
					if (lp && loop_comm_ances[lp->id][loop_map[dst->id]->id] == lp) {
						/* within the current loop */
						/* do not add loop header back to the queue */
						if (lpHeadId != dst->id) {
							p_enqueue(&pQueue, &(dst->id), dst->id);
							if (dbg) fprintf(dbgExec, "\n    Enqueued bbi %d", dst->id);
						}
					} else {// if (dst->id > lpTailId)
						/* outside the current loop */
						p_enqueue(outQueue, &(dst->id), dst->id);
						if (dbg) fprintf(dbgExec, "\n    OUTQUEUE enqueued bbi %d", dst->id);
					}
				}
            }
#endif
        }//end while queuing

        if (dbg) fprintf(dbgExec, "\n  pQueue is empty");
    }
    
    if (dbg && lp) fprintf(dbgExec,"\n----- Finish sym.exec of L%d -----",lp->id);
    
	return 1;
}

int doSymExec() {
    int dbg = 0;
    int i, j, k, m, id;
    inf_proc_t  *p;
    reg_t       *regList;
    inf_node_t  *ib;

    if (dbg) {
        printf("\nStart reg. expansion");fflush(stdout);
        fprintf(dbgExec,"\nStart reg. expansion");
    }

    /*perform register expansion*/
#ifdef _INF
    for( i = 0; i <prog.num_procs; i++ ) {
    	//printf("procedure:%d\n",i);
    	cntExec=0;
        p = &(inf_procs[i]);
        ib = &(p->inf_cfg[0]);
        regList = (reg_t*)(ib->regListIn);
        setInt(regList+0,0);
        setInt(regList+28,GLOBAL_START);
        setInt(regList+29,STACK_START);
        regUnknown(regList+30);

				loopSymExec(NULL,p,1,NULL);//perform symexec from outermost loop
        if (dbg) fprintf(dbgExec,"\nEnd of p: %d",i);fflush(dbgExec);
    }
#else /* sudiptac: perform register expansion on tcfg */
    	cntExec=0;
      regList = (reg_t*)(tcfg[0]->regListIn);
      setInt(regList+0,0);
#ifndef ARCH_ARM
      setInt(regList+28,GLOBAL_START);
      setInt(regList+29,STACK_START);
      regUnknown(regList+30);
#endif
			/* sudiptac: probably the tcfg[0] argument is obsolete,
			 * however, keeping it currently to maintain the
			 * compatibility */
			loopSymExec(loops[0],tcfg[0],1,NULL);
#endif

    if (dbg) {
        printf("\nEnd reg. expansion");fflush(stdout);
        fprintf(dbgExec,"\nEnd reg. expansion");
    }
}

static int initBlock() {
    int dbg=0;
    int i,j,k;
    inf_proc_t *p;
    inf_node_t *ib;
    reg_t *regList;

#ifdef _INF
    for( i = 0; i < prog.num_procs; i++ ) {
        p = &(inf_procs[i]);
        for( j = 0; j < p->num_bb; j++ ) {
            ib = &(p->inf_cfg[j]);
            ib->memListIn = NULL;
            ib->memListOut= NULL;
            //init regListIn
            ib->regListIn = calloc(NO_REG,sizeof(reg_t));
            initRegSet( (reg_t*)(ib->regListIn));
            //init regListOut
            ib->regListOut = calloc(NO_REG,sizeof(reg_t));
            initRegSet( (reg_t*)(ib->regListOut));
        }
    }
#else
    for( i = 0; i < num_tcfg_nodes; i++ ) {
            tcfg_node_t* bbi = tcfg[i];
            bbi->memListIn = NULL;
            bbi->memListOut= NULL;
            //init regListIn
            bbi->regListIn = calloc(NO_REG,sizeof(reg_t));
            initRegSet( (reg_t*)(bbi->regListIn));
            //init regListOut
            bbi->regListOut = calloc(NO_REG,sizeof(reg_t));
            initRegSet( (reg_t*)(bbi->regListOut));
    }
#endif
}
static void initAll() {
    dbgExec = fopen("dbg_exec.dbg","w");
    setDebugFile("dbg_comp.dbg");
    initBlock();

}
static void freeAll() {
    fclose(dbgExec);
}

int execute() {
    int dbg = 0;
    int i,j,k;
    inf_proc_t *p;
    inf_node_t *ib;
    reg_t *listIn, *listOut;

    ticks a,b;

    initAll();
    cntExec = 0;

    a= getticks();
    /* do symbolic execution to detect register with constant value */
    /* for detection of induction value, static memory address, etc */
    conflictDetectFlag = 0;
    traceInductionFlag = 0;
    traceMemFlag = 0;
    if (dbg) printf ("\nTrace constant value");
    doSymExec();
    if (dbg) printf ("\nFinish tracing constant value");

    /* reinitialize register value, keep known const*/
#ifdef _INF
    for (i=0; i<prog.num_procs; i++) {
        p= &(inf_procs[i]);
        for (j=0; j<p->num_bb; j++) {
            ib = &(p->inf_cfg[j]); 
            listIn = (reg_t*)(ib->regListIn);
            listOut= (reg_t*)(ib->regListOut);
            for (k=0; k<NO_REG; k++) {
                //if ( k == 28 || k == 29 ) continue;// || k == 30
                if (listIn[k].t!=VALUE_CONST) {
                    clrReg( listIn+k );
                    clrReg( listOut+k );
                }
            }
        }
    }
#else
    for (i=0; i<num_tcfg_nodes; i++) {
            tcfg_node_t* bbi = tcfg[i]; 
            listIn = (reg_t*)(bbi->regListIn);
            listOut= (reg_t*)(bbi->regListOut);
            
						for (k=0; k<NO_REG; k++) {
                //if ( k == 28 || k == 29 ) continue;// || k == 30
                if (listIn[k].t!=VALUE_CONST) {
                    clrReg( listIn+k );
                    clrReg( listOut+k );
                }
            }
    }
#endif
 
    /* detect induction value in register */
    if (dbg) printf ("\nDetect induction instruction");
    detectBIV();
    if (dbg) printf ("\nFinish detecting induction instruction");
    
    #if 1
    /* do symbolic execution again, now with induction var */
    /* NOTE: and with memory modeling if implemented */
    traceMemFlag = 1;
    traceInductionFlag = 1;
    if (dbg) printf("\nTrace mem & induction value");
    doSymExec();
    if (dbg) printf("\nFinish tracing mem & induction value");

    /* read symbolic execution result */
    if (dbg) {
        fprintf(dbgExec,"\nRead sym.exec result");fflush(dbgExec);
        printf("\nRead sym.exec result");fflush(stdout);
    }
    //conflictDetectFlag = 1;
#ifdef _INF
    for (i=0; i<prog.num_procs; i++) {
        p= &(inf_procs[i]);
        for (j = 0; j<p->num_bb; j++) {
            ib = &(p->inf_cfg[j]); 
            readSymExecResult(ib);
        }
    }
#else
    for (i = 0; i < num_tcfg_nodes; i++) {
        tcfg_node_t* bbi = tcfg[i]; 
        readSymExecResult(bbi);
    }
#endif
    #endif
    b= getticks();
	//printf("\n===================================================\n");
	//printf("Establish trace link %lf secs\n", (b - a)/((1.0) * CPU_MHZ));
	//printf("===================================================\n");

    freeAll();

    return 1;	
}
