/*******************************************************************************
 *
 * Chronos: A Timing Analyzer for Embedded Software
 * =============================================================================
 * http://www.comp.nus.edu.sg/~rpembed/chronos/
 *
 * Copyright (C) 2005 Xianfeng Li
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * $Id: tcfg.h,v 1.2 2006/06/24 08:54:57 lixianfe Exp $
 *
 * Construct a flow graph transformed from the original CFG due to function
 * inlining, loop unrolling ...
 *
 ******************************************************************************/

#ifndef TCFG_H
#define TCFG_H

#include "cfg.h"
#include "common.h"

struct task;
//typedef struct task *task_p;

// in many cases, two virtual nodes are assumed to exist: one is a "virtual
// start" node preceding the start node of the tcfg; the other is the "virtual
// end" node following end nodes of the tcfg
#define V_START_ID  -1
#define V_END_ID    -2
#define TCFG_STEP_SIZE 64

typedef struct tcfg_edge_t tcfg_edge_t;
typedef struct tcfg_node_t tcfg_node_t;

/*HBK: induction var for data cache*/
#define BIV_INST 0xa   /*is basic induction inst flag*/
#define OP_LEN 9       /* length of assembly token */
typedef struct { 
  char  addr[OP_LEN];
  char  op[OP_LEN];
  char  r1[OP_LEN];
  char  r2[OP_LEN];
  char  r3[OP_LEN];
  int   flag; /* mark special opr, e.g. induction operation */
  tcfg_node_t *bbi;		/* points to parent TCFG node */
} insn_t;

// CFG node type (basic block instance)
struct tcfg_node_t {
	cfg_node_t *bb; // pointer to the physical basic block
	int id; // global id in tcfg (has nothing to do with its bb id)
	tcfg_edge_t *in, *out; // incoming and outgoing edges
	unsigned flags;
	int loop_id; // vivy: for infeasible path constraints
	int exec_count; // vivy: for infeasible path constraints
	int anal_count;
	int dcache_delay; /* For data cache analysis */
	int inst_cache_delay; /* For instruction cache analysis using abstract
	 * interpretation */
	int n_persistence; /* Number of persistence instruction */
	int n_data_persistence; /* Number of persistence data blocks */
	int n_u1_persistence; /* Number of persistence instruction in Unified cache*/
	int n_u1_data_persistence; /* Number of persistence data blocks in unified cache */
	int n_l2_persistence; /* Number of instruction in l2 instruction cache */
	char is_unconditional;
	
	/*HBK: for cache analysis on tcfg*/
	int max_miss;
	/* worklist_p* acs_in;
	worklist_p* acs_out;*/
	acs_p** acs_in;
	acs_p** acs_out;

	acs_p ***must_inst;
	acs_p ***may_inst;
	acs_p ***may_inst_out;
	fifo_must_t **fifo_inst;
		
	fifo_must_t *fifo_in;
	fifo_must_t *fifo_out;
	int *mustdamage_out;
	int *mustdamage_in;

	worklist_p* mpa_acs_in;
	worklist_p* mpa_acs_out;
#if 0
	worklist_p mpa_acsin_arr, mpa_acsout_arr;
	worklist_p mpa_lp_in, mpa_lp_out;
	worklist_p mpa_acsl2in_arr, mpa_acsl2out_arr;
	worklist_p mpa_lpl2_in, mpa_lpl2_out;
#endif
	worklist_p* address_set;
	worklist_p* addrset_l1;
	worklist_p* addrset_l2;
	bitset_p dom_in, postdom_in;
	bitset_p dom_out, postdom_out;
	int equiv;
	int marked;
//	int *ucb;

  /* sudiptac: for making the address analysis 
   * inter-procedural */
  /* moving all the data related to the register 
   * expansion to tcfg node structure, as tcfg 
   * captures the inter-procedural CFG */
  /* register value tracing */
  void* regListIn;
  void* regListOut;
  /* memory value tracing */
  worklist_p memListIn;
  worklist_p memListOut;
  
  insn_t* insnlist;
    
  /*** HBK: for scope-aware data cache analysis ***/
  /* sudiptac: adding "d_instlist" here to make the 
   * address/data cache analysis inter-procedural */
  void   *d_instlist;

};

// since the numbers of outgoing/incoming edges of each tcfg node vary widely, a
// double link list is used, e.g., given any tcfg edge connecting src->dst, all
// outgoing edges of src can be traversed by going along two directions: going along
// prev_out will traverse all earlier out edges of src, and going along next_out will
// traverse all later out edges of src; similary for the incoming edges of dst
struct tcfg_edge_t {
	int id;
	int branch; // TAKEN or NOT_TAKEN
	tcfg_node_t *src, *dst;
	tcfg_edge_t *next_out; // next outgoing edge of src
	tcfg_edge_t *next_in; // next incoming edge of dst
	int flags;
	damage_p dmg;
	instdamage_p inst_dmg;
	char is_backedge;
};

typedef struct tcfg_nlink_t tcfg_nlink_t;
struct tcfg_nlink_t {
	tcfg_node_t *bbi;
	tcfg_nlink_t *next;
};

typedef struct tcfg_elink_t tcfg_elink_t;

struct tcfg_elink_t {
	tcfg_edge_t *edge;
	tcfg_elink_t *next;
};

void
prog_tran(int proc_id);

void
clear_bbi_flags();

int
bbi_type();

void
dump_tcfg(FILE *fp);

void
dump_visual_tcfg(FILE *fp);

int num_topo_tcfg;
int* topo_tcfg;
int num_topo_tcfg_loops;
int* topo_tcfg_loops;

/* dominance */ 
/*
int strictly_dominates(adnode_p, adnode_p);
int dominates(tcfg_node_t *n1, tcfg_node_t *n2);
void compute_dominance();
void print_dominance(void);
int strictly_postdominates(tcfg_node_t *n1, tcfg_node_t *n2);
int postdominates(tcfg_node_t *n1, tcfg_node_t *n2);
void compute_postdominance();
void print_postdominance(void);
void compute_cycle_equiv();
void print_cycle_equiv();
void build_tree();
*/

damage_p load_damage(FILE *f);

damage_p compute_all(worklist_p **miss_info);

void virtual_bypass(int src, int dst, damage_p dmg);
int ageing2crpd(worklist_p miss_list, int num_ageing);

tcfg_node_t **tcfg;

int num_tcfg_nodes, tcfg_size;

tcfg_edge_t **tcfg_edges;

int num_tcfg_edges;

tcfg_nlink_t ***bbi_map;

int bbi_pid(tcfg_node_t *bbi);
int bbi_bid(tcfg_node_t *bbi);
int cond_bbi(tcfg_node_t *bbi);
int bbi_backedge(tcfg_edge_t *edge);
void clear_tcfg_edge_flags();
int all_predecessors_visited(tcfg_node_t* bbi, char* visited);

#endif
